/***************************************************************************

	memory.c

	Functions which handle the CPU memory access.

	Caveats:

	* If your driver executes an opcode which crosses a bank-switched
	boundary, it will pull the wrong data out of memory. Although not
	a common case, you may need to revert to memcpy to work around this.
	See machine/tnzs.c for an example.

	To do:

	- Add fourth memory space for encrypted opcodes
	- Automatically mirror program space into data space if no data space
	- Get rid of association between memory_regions and RAM
	- Get rid of opcode/data separation by using address spaces?

***************************************************************************/

#include "driver.h"
#include "osd_cpu.h"
#include "state.h"

#include <stdarg.h>


#define MEM_DUMP		(0)



/***************************************************************************

	Basic theory of memory handling:

	An address with up to 32 bits is passed to a memory handler. First,
	an address mask is applied to the address, removing unused bits.

	Next, the address is broken into two halves, an upper half and a
	lower half. The number of bits in each half can be controlled via
	macros in memory.h, but they default to the upper 18 bits and the
	lower 14 bits. The upper half is then used as an index into the 
	base_lookup table.

	If the value pulled from the table is within the range 192-255, then
	the lower half of the address is needed to resolve the final handler.
	The value from the table (192-255) is combined with the lower address
	bits to form an index into a subtable.

	Table values in the range 0-31 are reserved for internal handling
	(such as RAM, ROM, NOP, and banking). Table values between 32 and 192
	are assigned dynamically at startup.

***************************************************************************/

/* macros for the profiler */
#define MEMREADSTART()			do { profiler_mark(PROFILER_MEMREAD); } while (0)
#define MEMREADEND(ret)			do { profiler_mark(PROFILER_END); return ret; } while (0)
#define MEMWRITESTART()			do { profiler_mark(PROFILER_MEMWRITE); } while (0)
#define MEMWRITEEND(ret)		do { (ret); profiler_mark(PROFILER_END); return; } while (0)

/* helper macros */
#define HANDLER_IS_RAM(h)		((FPTR)(h) == STATIC_RAM)
#define HANDLER_IS_ROM(h)		((FPTR)(h) == STATIC_ROM)
#define HANDLER_IS_RAMROM(h)	((FPTR)(h) == STATIC_RAMROM)
#define HANDLER_IS_NOP(h)		((FPTR)(h) == STATIC_NOP)
#define HANDLER_IS_BANK(h)		((FPTR)(h) >= STATIC_BANK1 && (FPTR)(h) <= STATIC_BANKMAX)
#define HANDLER_IS_STATIC(h)	((FPTR)(h) < STATIC_COUNT)

#define HANDLER_TO_BANK(h)		((FPTR)(h))
#define BANK_TO_HANDLER(b)		((void *)(b))

#define SPACE_SHIFT(s,a)		(((s)->ashift < 0) ? ((a) << -(s)->ashift) : ((a) >> (s)->ashift))
#define SPACE_SHIFT_END(s,a)	(((s)->ashift < 0) ? (((a) << -(s)->ashift) | ((1 << -(s)->ashift) - 1)) : ((a) >> (s)->ashift))
#define INV_SPACE_SHIFT(s,a)	(((s)->ashift < 0) ? ((a) >> -(s)->ashift) : ((a) << (s)->ashift))


/*-------------------------------------------------
	TYPE DEFINITIONS
-------------------------------------------------*/

struct bank_data
{
	UINT8 				used;				/* is this bank used? */
	UINT8 				dynamic;			/* is this bank allocated dynamically? */
	UINT8 				cpunum;				/* the CPU it is used for */
	UINT8 				spacenum;			/* the address space it is used for */
	offs_t 				base;				/* the base offset */
};

struct handler_data
{
	void *				handler;			/* function pointer for handler */
	offs_t				offset;				/* base offset for handler */
	offs_t				top;				/* maximum offset for handler */
};

struct table_data
{
	UINT8 *				table;				/* pointer to base of table */
	UINT8 				subtable_count;		/* number of subtables used */
	UINT8 				subtable_alloc;		/* number of subtables allocated */
	struct handler_data handlers[ENTRY_COUNT];/* array of user-installed handlers */
};

struct addrspace_data
{
	UINT8				cpunum;				/* CPU index */
	UINT8				spacenum;			/* address space index */
	INT8				ashift;				/* address shift */
	UINT8				abits;				/* address bits */
	UINT8 				dbits;				/* data bits */
	offs_t				mask;				/* address mask */
	data64_t			unmap;				/* unmapped value */
	struct table_data	read;				/* memory read lookup table */
	struct table_data	write;				/* memory write lookup table */
	struct memory_accessors_t *accessors;	/* pointer to the memory accessors */
};

struct cpu_data
{
	void *				rambase;			/* RAM base pointer */
	size_t				ramlength;			/* RAM length pointer */
	opbase_handler 		opbase;				/* opcode base handler */

	void *				op_ram;				/* dynamic RAM base pointer */
	void *				op_rom;				/* dynamic ROM base pointer */
	offs_t				op_mem_min;			/* dynamic ROM/RAM min */
	offs_t				op_mem_max;			/* dynamic ROM/RAM max */
	UINT8		 		opcode_entry;		/* opcode base handler */

	UINT8				spacemask;			/* mask of which address spaces are used */
	struct addrspace_data space[ADDRESS_SPACES]; /* info about each address space */
};


/*-------------------------------------------------
	GLOBAL VARIABLES
-------------------------------------------------*/

UINT8 *						OP_ROM;							/* opcode ROM base */
UINT8 *						OP_RAM;							/* opcode RAM base */
offs_t						OP_MEM_MIN;						/* opcode memory minimum */
offs_t						OP_MEM_MAX;						/* opcode memory maximum */
UINT8		 				opcode_entry;					/* opcode readmem entry */

struct address_space_t		address_space[ADDRESS_SPACES];	/* address space data */

UINT8 *						cpu_bankbase[STATIC_COUNT];		/* array of bank bases */
struct memory_block_t		ext_memory[MAX_EXT_MEMORY];		/* externally-allocated memory */
int 						ext_entries = 0;				/* number of entries ext_memory[] entries used */

static int					cur_context;					/* current CPU context */

static opbase_handler		opbasefunc;						/* opcode base override */

static struct cpu_data 		cpudata[MAX_CPU];				/* data gathered for each CPU */
static struct bank_data 	bankdata[MAX_BANKS+1];			/* data gathered for each bank */

static struct memory_accessors_t memory_accessors[ADDRESS_SPACES][4][2] =
{
	/* program accessors */
	{
		{
			{ program_read_byte_8, NULL, NULL, NULL, program_write_byte_8, NULL, NULL, NULL },
			{ program_read_byte_8, NULL, NULL, NULL, program_write_byte_8, NULL, NULL, NULL }
		},
		{
			{ program_read_byte_16le, program_read_word_16le, NULL, NULL, program_write_byte_16le, program_write_word_16le, NULL, NULL },
			{ program_read_byte_16be, program_read_word_16be, NULL, NULL, program_write_byte_16be, program_write_word_16be, NULL, NULL }
		},
		{
			{ program_read_byte_32le, program_read_word_32le, program_read_dword_32le, NULL, program_write_byte_32le, program_write_word_32le, program_write_dword_32le, NULL },
			{ program_read_byte_32be, program_read_word_32be, program_read_dword_32be, NULL, program_write_byte_32be, program_write_word_32be, program_write_dword_32be, NULL }
		},
		{
			{ program_read_byte_64le, program_read_word_64le, program_read_dword_64le, program_read_qword_64le, program_write_byte_64le, program_write_word_64le, program_write_dword_64le, program_write_qword_64le },
			{ program_read_byte_64be, program_read_word_64be, program_read_dword_64be, program_read_qword_64be, program_write_byte_64be, program_write_word_64be, program_write_dword_64be, program_write_qword_64be }
		}
	},
	
	/* data accessors */
	{
		{
			{ data_read_byte_8, NULL, NULL, NULL, data_write_byte_8, NULL, NULL, NULL },
			{ data_read_byte_8, NULL, NULL, NULL, data_write_byte_8, NULL, NULL, NULL }
		},
		{
			{ data_read_byte_16le, data_read_word_16le, NULL, NULL, data_write_byte_16le, data_write_word_16le, NULL, NULL },
			{ data_read_byte_16be, data_read_word_16be, NULL, NULL, data_write_byte_16be, data_write_word_16be, NULL, NULL }
		},
		{
			{ data_read_byte_32le, data_read_word_32le, data_read_dword_32le, NULL, data_write_byte_32le, data_write_word_32le, data_write_dword_32le, NULL },
			{ data_read_byte_32be, data_read_word_32be, data_read_dword_32be, NULL, data_write_byte_32be, data_write_word_32be, data_write_dword_32be, NULL }
		},
		{
			{ data_read_byte_64le, data_read_word_64le, data_read_dword_64le, data_read_qword_64le, data_write_byte_64le, data_write_word_64le, data_write_dword_64le, data_write_qword_64le },
			{ data_read_byte_64be, data_read_word_64be, data_read_dword_64be, data_read_qword_64be, data_write_byte_64be, data_write_word_64be, data_write_dword_64be, data_write_qword_64be }
		}
	},
	
	/* I/O accessors */
	{
		{
			{ io_read_byte_8, NULL, NULL, NULL, io_write_byte_8, NULL, NULL, NULL },
			{ io_read_byte_8, NULL, NULL, NULL, io_write_byte_8, NULL, NULL, NULL }
		},
		{
			{ io_read_byte_16le, io_read_word_16le, NULL, NULL, io_write_byte_16le, io_write_word_16le, NULL, NULL },
			{ io_read_byte_16be, io_read_word_16be, NULL, NULL, io_write_byte_16be, io_write_word_16be, NULL, NULL }
		},
		{
			{ io_read_byte_32le, io_read_word_32le, io_read_dword_32le, NULL, io_write_byte_32le, io_write_word_32le, io_write_dword_32le, NULL },
			{ io_read_byte_32be, io_read_word_32be, io_read_dword_32be, NULL, io_write_byte_32be, io_write_word_32be, io_write_dword_32be, NULL }
		},
		{
			{ io_read_byte_64le, io_read_word_64le, io_read_dword_64le, io_read_qword_64le, io_write_byte_64le, io_write_word_64le, io_write_dword_64le, io_write_qword_64le },
			{ io_read_byte_64be, io_read_word_64be, io_read_dword_64be, io_read_qword_64be, io_write_byte_64be, io_write_word_64be, io_write_dword_64be, io_write_qword_64be }
		}
	},
};



/*-------------------------------------------------
	PROTOTYPES
-------------------------------------------------*/

static int CLIB_DECL fatalerror(const char *string, ...);
static int init_cpudata(void);
static int init_addrspace(UINT8 cpunum, UINT8 spacenum);
static int verify_memory(void);
static int populate_memory(void);
static void install_mem_handler(struct addrspace_data *space, int iswrite, int databits, offs_t start, offs_t end, offs_t mask, void *handler);
static void *assign_dynamic_bank(int cpunum, int spacenum, offs_t start);
static UINT8 get_handler_index(struct handler_data *table, void *handler, offs_t start);
static void populate_table(struct addrspace_data *space, int iswrite, offs_t start, offs_t stop, UINT8 handler);
static UINT8 alloc_new_subtable(const struct addrspace_data *space, struct table_data *tabledata, UINT8 previous_value);
static int allocate_memory(void);
static int allocate_memory_block(int cpunum, int spacenum, offs_t start, offs_t end);
static void register_for_save(int cpunum, int spacenum, offs_t start, void *base, size_t numbytes);
static int find_memory(void);
static void *memory_find_base(int cpunum, int spacenum, offs_t offset);
static void *get_static_handler(int databits, int readorwrite, int spacenum, int which);

#if (MEM_DUMP)
static void mem_dump(void);
#else
#define mem_dump()
#endif



/*-------------------------------------------------
	memory_init - initialize the memory system
-------------------------------------------------*/

int memory_init(void)
{
	/* no current context to start */
	cur_context = -1;

	/* init the CPUs */
	if (!init_cpudata())
		return 0;

	/* verify the memory handlers and check banks */
	if (!verify_memory())
		return 0;

	/* then fill in the tables */
	if (!populate_memory())
		return 0;

	/* allocate any necessary memory */
	if (!allocate_memory())
		return 0;

	/* find all the allocated pointers */
	if (!find_memory())
		return 0;

	/* dump the final memory configuration */
	mem_dump();
	return 1;
}


/*-------------------------------------------------
	memory_shutdown - free memory
-------------------------------------------------*/

void memory_shutdown(void)
{
	int cpunum, spacenum, extnum;
	struct memory_block_t *ext;

	/* free all the tables */
	for (cpunum = 0; cpunum < MAX_CPU; cpunum++)
		for (spacenum = 0; spacenum < ADDRESS_SPACES; spacenum++)
		{
			if (cpudata[cpunum].space[spacenum].read.table)
				free(cpudata[cpunum].space[spacenum].read.table);
			if (cpudata[cpunum].space[spacenum].write.table)
				free(cpudata[cpunum].space[spacenum].write.table);
		}
	memset(&cpudata, 0, sizeof(cpudata));

	/* free all the external memory */
	ext = ext_memory;
	for (extnum = 0; extnum < ext_entries; extnum++)
		free(ext_memory[extnum].data);
	memset(ext_memory, 0, sizeof(ext_memory));
	ext_entries = 0;
}


/*-------------------------------------------------
	memory_set_context - set the memory context
-------------------------------------------------*/

void memory_set_context(int activecpu)
{
	/* remember dynamic RAM/ROM */
	if (cur_context != -1)
	{
		cpudata[cur_context].op_ram = OP_RAM;
		cpudata[cur_context].op_rom = OP_ROM;
		cpudata[cur_context].op_mem_min = OP_MEM_MIN;
		cpudata[cur_context].op_mem_max = OP_MEM_MAX;
		cpudata[cur_context].opcode_entry = opcode_entry;
	}
	cur_context = activecpu;

	cpu_bankbase[STATIC_RAM] = cpudata[activecpu].rambase;
	OP_RAM = cpudata[activecpu].op_ram;
	OP_ROM = cpudata[activecpu].op_rom;
	OP_MEM_MIN = cpudata[activecpu].op_mem_min;
	OP_MEM_MAX = cpudata[activecpu].op_mem_max;
	opcode_entry = cpudata[activecpu].opcode_entry;

	/* program address space */
	address_space[ADDRESS_SPACE_PROGRAM].addrmask = cpudata[activecpu].space[ADDRESS_SPACE_PROGRAM].mask;
	address_space[ADDRESS_SPACE_PROGRAM].readlookup = cpudata[activecpu].space[ADDRESS_SPACE_PROGRAM].read.table;
	address_space[ADDRESS_SPACE_PROGRAM].writelookup = cpudata[activecpu].space[ADDRESS_SPACE_PROGRAM].write.table;
	address_space[ADDRESS_SPACE_PROGRAM].readhandlers = cpudata[activecpu].space[ADDRESS_SPACE_PROGRAM].read.handlers;
	address_space[ADDRESS_SPACE_PROGRAM].writehandlers = cpudata[activecpu].space[ADDRESS_SPACE_PROGRAM].write.handlers;
	address_space[ADDRESS_SPACE_PROGRAM].accessors = cpudata[activecpu].space[ADDRESS_SPACE_PROGRAM].accessors;

	/* data address space */
	if (cpudata[activecpu].spacemask & (1 << ADDRESS_SPACE_DATA))
	{
		address_space[ADDRESS_SPACE_DATA].addrmask = cpudata[activecpu].space[ADDRESS_SPACE_DATA].mask;
		address_space[ADDRESS_SPACE_DATA].readlookup = cpudata[activecpu].space[ADDRESS_SPACE_DATA].read.table;
		address_space[ADDRESS_SPACE_DATA].writelookup = cpudata[activecpu].space[ADDRESS_SPACE_DATA].write.table;
		address_space[ADDRESS_SPACE_DATA].readhandlers = cpudata[activecpu].space[ADDRESS_SPACE_DATA].read.handlers;
		address_space[ADDRESS_SPACE_DATA].writehandlers = cpudata[activecpu].space[ADDRESS_SPACE_DATA].write.handlers;
		address_space[ADDRESS_SPACE_DATA].accessors = cpudata[activecpu].space[ADDRESS_SPACE_DATA].accessors;
	}

	/* I/O address space */
	if (cpudata[activecpu].spacemask & (1 << ADDRESS_SPACE_IO))
	{
		address_space[ADDRESS_SPACE_IO].addrmask = cpudata[activecpu].space[ADDRESS_SPACE_IO].mask;
		address_space[ADDRESS_SPACE_IO].readlookup = cpudata[activecpu].space[ADDRESS_SPACE_IO].read.table;
		address_space[ADDRESS_SPACE_IO].writelookup = cpudata[activecpu].space[ADDRESS_SPACE_IO].write.table;
		address_space[ADDRESS_SPACE_IO].readhandlers = cpudata[activecpu].space[ADDRESS_SPACE_IO].read.handlers;
		address_space[ADDRESS_SPACE_IO].writehandlers = cpudata[activecpu].space[ADDRESS_SPACE_IO].write.handlers;
		address_space[ADDRESS_SPACE_IO].accessors = cpudata[activecpu].space[ADDRESS_SPACE_IO].accessors;
	}

	opbasefunc = cpudata[activecpu].opbase;
}


/*-------------------------------------------------
	memory_set_opbase_handler - change op-code
	memory base
-------------------------------------------------*/

opbase_handler memory_set_opbase_handler(int cpunum, opbase_handler function)
{
	opbase_handler old = cpudata[cpunum].opbase;
	cpudata[cpunum].opbase = function;
	if (cpunum == cpu_getactivecpu())
		opbasefunc = function;
	return old;
}


/*-------------------------------------------------
	memory_set_opbase - generic opcode base changer
-------------------------------------------------*/

void memory_set_opbase(offs_t pc)
{
	UINT8 *base;
	UINT8 entry;

	/* allow overrides */
	if (opbasefunc)
	{
		pc = (*opbasefunc)(pc);
		if (pc == ~0)
			return;
	}

	/* perform the lookup */
	pc &= address_space[ADDRESS_SPACE_PROGRAM].addrmask;
	entry = address_space[ADDRESS_SPACE_PROGRAM].readlookup[LEVEL1_INDEX(pc)];
	if (entry >= SUBTABLE_BASE)
		entry = address_space[ADDRESS_SPACE_PROGRAM].readlookup[LEVEL2_INDEX(entry,pc)];
	opcode_entry = entry;

	/* RAM/ROM/RAMROM */
	if (entry >= STATIC_RAM && entry <= STATIC_RAMROM)
		base = cpu_bankbase[STATIC_RAM];

	/* banked memory */
	else if (entry >= STATIC_BANK1 && entry <= STATIC_RAM)
		base = cpu_bankbase[entry];

	/* other memory -- could be very slow! */
	else
	{
		logerror("cpu #%d (PC=%08X): warning - op-code execute on mapped I/On",
					cpu_getactivecpu(), activecpu_get_pc());
		/*base = memory_find_base(cpu_getactivecpu(), pc);*/
		return;
	}

	/* compute the adjusted base */
	OP_ROM = base - address_space[ADDRESS_SPACE_PROGRAM].readhandlers[entry].offset + (OP_ROM - OP_RAM);
	OP_RAM = base - address_space[ADDRESS_SPACE_PROGRAM].readhandlers[entry].offset;
	OP_MEM_MIN = address_space[ADDRESS_SPACE_PROGRAM].readhandlers[entry].offset;
	OP_MEM_MAX = (entry >= STATIC_RAM && entry <= STATIC_RAMROM)
		? cpudata[cpu_getactivecpu()].ramlength - 1
		: address_space[ADDRESS_SPACE_PROGRAM].readhandlers[entry].top;
}


/*-------------------------------------------------
	memory_set_opcode_base - set the base of
	ROM
-------------------------------------------------*/

void memory_set_opcode_base(int cpunum, void *base)
{
	if (cur_context == cpunum)
	{
		OP_ROM = base;
		OP_MEM_MIN = (offs_t) 0x00000000;
		OP_MEM_MAX = (offs_t) 0x7fffffff;
	}
	else
	{
		cpudata[cpunum].op_rom = base;
		cpudata[cpunum].op_mem_min = (offs_t) 0x00000000;
		cpudata[cpunum].op_mem_max = (offs_t) 0x7fffffff;
	}
}


/*-------------------------------------------------
	memory_get_read_ptr - return a pointer to the
	base of RAM associated with the given CPU
	and offset
-------------------------------------------------*/

void *memory_get_read_ptr(int cpunum, int spacenum, offs_t offset)
{
	struct addrspace_data *space = &cpudata[cpunum].space[spacenum];
	UINT8 entry;

	/* perform the lookup */
	offset &= space->mask;
	entry = space->read.table[LEVEL1_INDEX(offset)];
	if (entry >= SUBTABLE_BASE)
		entry = space->read.table[LEVEL2_INDEX(entry, offset)];

	/* 8-bit case: RAM/ROM/RAMROM */
	if (entry > STATIC_RAM)
		return NULL;
	offset -= space->read.handlers[entry].offset;
	return &cpu_bankbase[entry][offset];
}


/*-------------------------------------------------
	memory_get_write_ptr - return a pointer to the
	base of RAM associated with the given CPU
	and offset
-------------------------------------------------*/

void *memory_get_write_ptr(int cpunum, int spacenum, offs_t offset)
{
	struct addrspace_data *space = &cpudata[cpunum].space[spacenum];
	UINT8 entry;

	/* perform the lookup */
	offset &= space->mask;
	entry = space->write.table[LEVEL1_INDEX(offset)];
	if (entry >= SUBTABLE_BASE)
		entry = space->write.table[LEVEL2_INDEX(entry, offset)];

	/* 8-bit case: RAM/ROM/RAMROM */
	if (entry > STATIC_RAM)
		return NULL;
	offset -= space->write.handlers[entry].offset;
	return &cpu_bankbase[entry][offset];
}


/*-------------------------------------------------
	memory_install_readX_handler - install dynamic
	read handler for X-bit case
-------------------------------------------------*/

data8_t *memory_install_read8_handler(int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, read8_handler handler)
{
	struct addrspace_data *space = &cpudata[cpunum].space[spacenum];
	install_mem_handler(space, 0, 8, start, end, mask, (void *)handler);
	mem_dump();
	return memory_find_base(cpunum, spacenum, SPACE_SHIFT(space, start));
}

data16_t *memory_install_read16_handler(int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, read16_handler handler)
{
	struct addrspace_data *space = &cpudata[cpunum].space[spacenum];
	install_mem_handler(space, 0, 16, start, end, mask, (void *)handler);
	mem_dump();
	return memory_find_base(cpunum, spacenum, SPACE_SHIFT(space, start));
}

data32_t *memory_install_read32_handler(int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, read32_handler handler)
{
	struct addrspace_data *space = &cpudata[cpunum].space[spacenum];
	install_mem_handler(space, 0, 32, start, end, mask, (void *)handler);
	mem_dump();
	return memory_find_base(cpunum, spacenum, SPACE_SHIFT(space, start));
}

data64_t *memory_install_read64_handler(int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, read64_handler handler)
{
	struct addrspace_data *space = &cpudata[cpunum].space[spacenum];
	install_mem_handler(space, 0, 64, start, end, mask, (void *)handler);
	mem_dump();
	return memory_find_base(cpunum, spacenum, SPACE_SHIFT(space, start));
}


/*-------------------------------------------------
	memory_install_writeX_handler - install dynamic
	write handler for X-bit case
-------------------------------------------------*/

data8_t *memory_install_write8_handler(int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, write8_handler handler)
{
	struct addrspace_data *space = &cpudata[cpunum].space[spacenum];
	install_mem_handler(space, 1, 8, start, end, mask, (void *)handler);
	mem_dump();
	return memory_find_base(cpunum, spacenum, SPACE_SHIFT(space, start));
}

data16_t *memory_install_write16_handler(int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, write16_handler handler)
{
	struct addrspace_data *space = &cpudata[cpunum].space[spacenum];
	install_mem_handler(space, 1, 16, start, end, mask, (void *)handler);
	mem_dump();
	return memory_find_base(cpunum, spacenum, SPACE_SHIFT(space, start));
}

data32_t *memory_install_write32_handler(int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, write32_handler handler)
{
	struct addrspace_data *space = &cpudata[cpunum].space[spacenum];
	install_mem_handler(space, 1, 32, start, end, mask, (void *)handler);
	mem_dump();
	return memory_find_base(cpunum, spacenum, SPACE_SHIFT(space, start));
}

data64_t *memory_install_write64_handler(int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, write64_handler handler)
{
	struct addrspace_data *space = &cpudata[cpunum].space[spacenum];
	install_mem_handler(space, 1, 64, start, end, mask, (void *)handler);
	mem_dump();
	return memory_find_base(cpunum, spacenum, SPACE_SHIFT(space, start));
}


/*-------------------------------------------------
	fatalerror - display an error message and
	exit immediately
-------------------------------------------------*/

static int CLIB_DECL fatalerror(const char *string, ...)
{
	va_list arg;
	va_start(arg, string);
	vprintf(string, arg);
	va_end(arg);
	exit(1);
	return 0;
}


/*-------------------------------------------------
	init_cpudata - initialize the cpudata
	structure for each CPU
-------------------------------------------------*/

static int init_cpudata(void)
{
	int cpunum, spacenum;

	/* zap the cpudata structure */
	memset(&cpudata, 0, sizeof(cpudata));

	/* loop over CPUs */
	for (cpunum = 0; cpunum < cpu_gettotalcpu(); cpunum++)
	{
		int cputype = Machine->drv->cpu[cpunum].cpu_type;

		/* set the RAM/ROM base */
		cpudata[cpunum].rambase = cpudata[cpunum].op_ram = cpudata[cpunum].op_rom = memory_region(REGION_CPU1 + cpunum);
		cpudata[cpunum].op_mem_max = cpudata[cpunum].ramlength = memory_region_length(REGION_CPU1 + cpunum);
		cpudata[cpunum].op_mem_min = 0;
		cpudata[cpunum].opcode_entry = STATIC_ROM;
		cpudata[cpunum].opbase = NULL;

		/* TODO: make this dynamic */
		cpudata[cpunum].spacemask = 0;
		for (spacenum = 0; spacenum < ADDRESS_SPACES; spacenum++)
			if (!init_addrspace(cpunum, spacenum))
				return 0;
		
		/* Z80 port mask kludge */
		if (cputype == CPU_Z80 || cputype == CPU_Z180)
			if (!(Machine->drv->cpu[cpunum].cpu_flags & CPU_16BIT_PORT))
				cpudata[cpunum].space[ADDRESS_SPACE_IO].mask = 0xff;
	}
	return 1;
}


/*-------------------------------------------------
	init_addrspace - initialize the address space
	data structure
-------------------------------------------------*/

static int init_addrspace(UINT8 cpunum, UINT8 spacenum)
{
	struct addrspace_data *space = &cpudata[cpunum].space[spacenum];
	int cputype = Machine->drv->cpu[cpunum].cpu_type;
	int abits = cputype_addrbus_width(cputype, spacenum);
	int dbits = cputype_databus_width(cputype, spacenum);
	int accessorindex = (dbits == 8) ? 0 : (dbits == 16) ? 1 : (dbits == 32) ? 2 : 3;
	int entrynum;

	/* determine the address and data bits */
	space->cpunum = cpunum;
	space->spacenum = spacenum;
	space->ashift = cputype_addrbus_shift(cputype, spacenum);
	space->abits = abits - space->ashift;
	space->dbits = dbits;
	space->mask = SPACE_SHIFT(space, 0xffffffffUL >> (32 - abits));
	space->accessors = &memory_accessors[spacenum][accessorindex][cputype_endianness(cputype) == CPU_IS_LE ? 0 : 1];
	
	/* if there's nothing here, just punt */
	if (space->abits == 0)
		return 1;
	cpudata[cpunum].spacemask |= 1 << spacenum;

	/* init the static handlers */
	memset(space->read.handlers, 0, sizeof(space->read.handlers));
	memset(space->write.handlers, 0, sizeof(space->write.handlers));
	for (entrynum = 0; entrynum < ENTRY_COUNT; entrynum++)
	{
		space->read.handlers[entrynum].handler = get_static_handler(dbits, 0, spacenum, entrynum);
		space->write.handlers[entrynum].handler = get_static_handler(dbits, 1, spacenum, entrynum);
	}

	/* allocate memory */
	space->read.table = malloc(1 << LEVEL1_BITS);
	space->write.table = malloc(1 << LEVEL1_BITS);
	if (!space->read.table)
		return fatalerror("cpu #%d couldn't allocate read table\n", cpunum);
	if (!space->write.table)
		return fatalerror("cpu #%d couldn't allocate write table\n", cpunum);

	/* initialize everything to unmapped */
	memset(space->read.table, STATIC_UNMAP, 1 << LEVEL1_BITS);
	memset(space->write.table, STATIC_UNMAP, 1 << LEVEL1_BITS);
	return 1;
}


/*-------------------------------------------------
	verify_memory - verify the memory structs
	and track which banks are referenced
-------------------------------------------------*/

static int verify_memory(void)
{
	int cpunum, spacenum;

	/* zap the bank data */
	memset(&bankdata, 0, sizeof(bankdata));

	/* loop over CPUs */
	for (cpunum = 0; cpunum < cpu_gettotalcpu(); cpunum++)
		for (spacenum = 0; spacenum < ADDRESS_SPACES; spacenum++)
		{
			const struct memory_read8_map_t *mra = Machine->drv->cpu[cpunum].memory_map[spacenum].read;
			const struct memory_write8_map_t *mwa = Machine->drv->cpu[cpunum].memory_map[spacenum].write;
			struct addrspace_data *space = &cpudata[cpunum].space[spacenum];
			UINT32 width;
			int bank;

			/* determine the desired width */
			switch (space->dbits)
			{
				case 8:		width = MEMORY_WIDTH_8;		break;
				case 16:	width = MEMORY_WIDTH_16;	break;
				case 32:	width = MEMORY_WIDTH_32;	break;
				case 64:	width = MEMORY_WIDTH_64;	break;
				case 0:		continue;
				default:	return fatalerror("cpu #%d has invalid memory width for address space %d!\n", cpunum, spacenum);
			}

			/* verify the read handlers */
			if (mra)
			{
				/* verify the PROGRAM_READ8_MAP_START header */
				if (mra->start == MEMORY_MARKER && mra->end != 0)
				{
					if (((mra->end & MEMORY_SPACE_MASK) >> MEMORY_SPACE_SHIFT) != spacenum + 1)
						return fatalerror("cpu #%d has address space %d read handlers in place of address space %d read handlers!\n", cpunum, ((mra->end & MEMORY_SPACE_MASK) >> MEMORY_SPACE_SHIFT) - 1, spacenum);
					if ((mra->end & MEMORY_DIRECTION_MASK) != MEMORY_DIRECTION_READ)
						return fatalerror("cpu #%d has address space %d write handlers in place of address space %d read handlers!\n", cpunum, spacenum, spacenum);
					if ((mra->end & MEMORY_WIDTH_MASK) != width)
						return fatalerror("cpu #%d uses wrong data width read handlers for address space %d! (width = %d, memory = %08x)\n", cpunum, spacenum, cpunum_databus_width(cpunum, spacenum), mra->end);
					mra++;
				}

				/* track banks used */
				for ( ; !IS_MEMORY_END(mra); mra++)
					if (!IS_MEMORY_MARKER(mra) && HANDLER_IS_BANK(mra->handler))
					{
						bank = HANDLER_TO_BANK(mra->handler);
						bankdata[bank].used = 1;
						bankdata[bank].dynamic = 0;
						bankdata[bank].cpunum = cpunum;
						bankdata[bank].spacenum = spacenum;
						bankdata[bank].base = SPACE_SHIFT(space, mra->start);
					}
			}

			/* verify the write handlers */
			if (mwa)
			{
				/* verify the PROGRAM_WRITE8_MAP_START header */
				if (mwa->start == MEMORY_MARKER && mwa->end != 0)
				{
					if (((mwa->end & MEMORY_SPACE_MASK) >> MEMORY_SPACE_SHIFT) != spacenum + 1)
						return fatalerror("cpu #%d has address space %d write handlers in place of address space %d write handlers!\n", cpunum, ((mwa->end & MEMORY_SPACE_MASK) >> MEMORY_SPACE_SHIFT) - 1, spacenum);
					if ((mwa->end & MEMORY_DIRECTION_MASK) != MEMORY_DIRECTION_WRITE)
						return fatalerror("cpu #%d has address space %d read handlers in place of address space %d write handlers!\n", cpunum, spacenum, spacenum);
					if ((mwa->end & MEMORY_WIDTH_MASK) != width)
						return fatalerror("cpu #%d uses wrong data width write handlers for address space %d! (width = %d, memory = %08x)\n", cpunum, spacenum, cpunum_databus_width(cpunum, spacenum), mwa->end);
					mwa++;
				}

				/* track banks used */
				for (; !IS_MEMORY_END(mwa); mwa++)
					if (!IS_MEMORY_MARKER(mwa) && HANDLER_IS_BANK(mwa->handler))
					{
						bank = HANDLER_TO_BANK(mwa->handler);
						bankdata[bank].used = 1;
						bankdata[bank].dynamic = 0;
						bankdata[bank].cpunum = cpunum;
						bankdata[bank].spacenum = spacenum;
						bankdata[bank].base = SPACE_SHIFT(space, mwa->start);
					}
			}
		}
	return 1;
}


/*-------------------------------------------------
	populate_memory - populate the memory mapping
	tables with entries
-------------------------------------------------*/

static int populate_memory(void)
{
	int cpunum, spacenum;

	/* loop over CPUs and address spaces */
	for (cpunum = 0; cpunum < cpu_gettotalcpu(); cpunum++)
		for (spacenum = 0; spacenum < ADDRESS_SPACES; spacenum++)
			if (cpudata[cpunum].spacemask & (1 << spacenum))
			{
				const struct memory_read8_map_t *mra, *mra_start = Machine->drv->cpu[cpunum].memory_map[spacenum].read;
				const struct memory_write8_map_t *mwa, *mwa_start = Machine->drv->cpu[cpunum].memory_map[spacenum].write;
				struct addrspace_data *space = &cpudata[cpunum].space[spacenum];
				
				/* install the read handlers */
				if (mra_start)
				{
					/* first find the end and check for special values */
					for (mra = mra_start; !IS_MEMORY_END(mra); mra++)
						if (IS_MEMORY_MARKER(mra))
						{
							if (mra->end & MEMORY_ABITS_MASK)
								space->mask = 0xffffffffUL >> (32 - ((mra->end & MEMORY_ABITS_MASK) >> MEMORY_ABITS_SHIFT));
							if (mra->end & MEMORY_UNMAPPED_MASK)
								space->unmap = ((mra->end & MEMORY_UNMAPPED_MASK) == MEMORY_UNMAPPED_0) ? (data64_t)0 : (data64_t)-1;
						}

					/* then work backwards */
					for (mra--; mra >= mra_start; mra--)
						if (!IS_MEMORY_MARKER(mra))
							install_mem_handler(space, 0, space->dbits, mra->start, mra->end, mra->mask, (void *)mra->handler);
				}

				/* install the write handlers */
				if (mwa_start)
				{
					/* first find the end and check for address bits */
					for (mwa = mwa_start; !IS_MEMORY_END(mwa); mwa++)
						if (IS_MEMORY_MARKER(mwa))
						{
							if (mwa->end & MEMORY_ABITS_MASK)
								space->mask = 0xffffffffUL >> (32 - ((mwa->end & MEMORY_ABITS_MASK) >> MEMORY_ABITS_SHIFT));
							if (mwa->end & MEMORY_UNMAPPED_MASK)
								space->unmap = ((mwa->end & MEMORY_UNMAPPED_MASK) == MEMORY_UNMAPPED_0) ? (data64_t)0 : (data64_t)-1;
						}

					/* then work backwards */
					for (mwa--; mwa >= mwa_start; mwa--)
						if (!IS_MEMORY_MARKER(mwa))
							install_mem_handler(space, 1, space->dbits, mwa->start, mwa->end, mwa->mask, (void *)mwa->handler);
				}
			}

	return 1;
}


/*-------------------------------------------------
	install_mem_handler - installs a handler for
	memory operations
-------------------------------------------------*/

static void install_mem_handler(struct addrspace_data *space, int iswrite, int databits, offs_t start, offs_t end, offs_t mask, void *handler)
{
	struct table_data *tabledata = iswrite ? &space->write : &space->read;
	UINT8 idx;

	/* sanity check */
	if (space->dbits != databits)
	{
		printf("fatal: install_mem_handler called with a %d-bit handler for a %d-bit address space\n", databits, space->dbits);
		exit(1);
	}

	/* adjust the addresses for the shift value */
	start = SPACE_SHIFT(space, start);
	end = SPACE_SHIFT_END(space, end);
	mask = SPACE_SHIFT(space, mask);

	/* translate ROM and RAMROM to RAM here for read cases */
	if (!iswrite)
		if (HANDLER_IS_ROM(handler) || HANDLER_IS_RAMROM(handler))
			handler = (void *)MRA8_RAM;

	/* assign banks for sparse memory spaces */
	if ((space->spacenum != ADDRESS_SPACE_PROGRAM || IS_SPARSE(space->abits)) && HANDLER_IS_RAM(handler))
		handler = (void *)assign_dynamic_bank(space->cpunum, space->spacenum, start);

	/* set the handler */
	idx = get_handler_index(tabledata->handlers, handler, start);
	populate_table(space, iswrite, start, end, idx);

	/* if this is being installed to a live CPU, update the context */
	if (space->cpunum == cur_context)
		memory_set_context(cur_context);
}


/*-------------------------------------------------
	assign_dynamic_bank - finds a free or exact
	matching bank
-------------------------------------------------*/

static void *assign_dynamic_bank(int cpunum, int spacenum, offs_t start)
{
	int bank;

	/* special case: never assign a dynamic bank to an offset that */
	/* intersects the CPU's region; always use RAM for that */
	if (spacenum == ADDRESS_SPACE_PROGRAM && start < memory_region_length(REGION_CPU1 + cpunum))
		return (void *)STATIC_RAM;

	/* loop over banks, searching for an exact match or an empty */
	for (bank = 1; bank <= MAX_BANKS; bank++)
		if (!bankdata[bank].used || (bankdata[bank].dynamic && bankdata[bank].cpunum == cpunum && bankdata[bank].spacenum == spacenum && bankdata[bank].base == start))
		{
			bankdata[bank].used = 1;
			bankdata[bank].dynamic = 1;
			bankdata[bank].cpunum = cpunum;
			bankdata[bank].spacenum = spacenum;
			bankdata[bank].base = start;
			return BANK_TO_HANDLER(bank);
		}

	/* if we got here, we failed */
	fatalerror("cpu #%d: ran out of banks for sparse memory regions!\n", cpunum);
	return NULL;
}


/*-------------------------------------------------
	get_handler_index - finds the index of a
	handler, or allocates a new one as necessary
-------------------------------------------------*/

static UINT8 get_handler_index(struct handler_data *table, void *handler, offs_t start)
{
	int i;

	/* all static handlers are hardcoded */
	if (HANDLER_IS_STATIC(handler))
		return (FPTR)handler;

	/* otherwise, we have to search */
	for (i = STATIC_COUNT; i < SUBTABLE_BASE; i++)
	{
		if (table[i].handler == NULL)
		{
			table[i].handler = handler;
			table[i].offset = start;
		}
		if (table[i].handler == handler && table[i].offset == start)
			return i;
	}
	return 0;
}


/*-------------------------------------------------
	populate_table - assign a memory handler to
	a range of addresses
-------------------------------------------------*/

static void populate_table(struct addrspace_data *space, int iswrite, offs_t start, offs_t stop, UINT8 handler)
{
	struct table_data *tabledata = iswrite ? &space->write : &space->read;
	offs_t l2mask = (1 << LEVEL2_BITS) - 1;
	offs_t l1start = start >> LEVEL2_BITS;
	offs_t l2start = start & l2mask;
	offs_t l1stop = stop >> LEVEL2_BITS;
	offs_t l2stop = stop & l2mask;
	offs_t l1index;
	UINT8 subindex;

	/* sanity check */
	if (start > stop)
		return;

	/* set the base for non RAM/ROM cases */
	if (handler != STATIC_RAM && handler != STATIC_ROM && handler != STATIC_RAMROM)
	{
		tabledata->handlers[handler].offset = start;
		tabledata->handlers[handler].top = stop;
	}

	/* handle the starting edge if it's not on a block boundary */
	if (l2start != 0)
	{
		/* get the subtable index */
		subindex = tabledata->table[l1start];
		if (subindex < SUBTABLE_BASE)
			subindex = tabledata->table[l1start] = alloc_new_subtable(space, tabledata, subindex);
		subindex &= SUBTABLE_MASK;

		/* if the start and stop end within the same block, handle that */
		if (l1start == l1stop)
		{
			memset(&tabledata->table[(1 << LEVEL1_BITS) + (subindex << LEVEL2_BITS) + l2start], handler, l2stop - l2start + 1);
			return;
		}

		/* otherwise, fill until the end */
		memset(&tabledata->table[(1 << LEVEL1_BITS) + (subindex << LEVEL2_BITS) + l2start], handler, (1 << LEVEL2_BITS) - l2start);
		if (l1start != (offs_t)~0) l1start++;
	}

	/* handle the trailing edge if it's not on a block boundary */
	if (l2stop != l2mask)
	{
		/* get the subtable index */
		subindex = tabledata->table[l1stop];
		if (subindex < SUBTABLE_BASE)
			subindex = tabledata->table[l1stop] = alloc_new_subtable(space, tabledata, subindex);
		subindex &= SUBTABLE_MASK;

		/* fill from the beginning */
		memset(&tabledata->table[(1 << LEVEL1_BITS) + (subindex << LEVEL2_BITS)], handler, l2stop + 1);

		/* if the start and stop end within the same block, handle that */
		if (l1start == l1stop)
			return;
		if (l1stop != 0) l1stop--;
	}

	/* now fill in the middle tables */
	for (l1index = l1start; l1index <= l1stop; l1index++)
	{
		/* if we don't have a subtable here. just set the L1 value */
		if (tabledata->table[l1index] < SUBTABLE_BASE)
			tabledata->table[l1index] = handler;
		
		/* otherwise, fill in the full subtable since we don't have a means to reclaim "deallocated" subtables */
		else
			memset(&tabledata->table[(1 << LEVEL1_BITS) + ((tabledata->table[l1index] & SUBTABLE_MASK) << LEVEL2_BITS)], handler, 1 << LEVEL2_BITS);
	}
}


/*-------------------------------------------------
	alloc_new_subtable - allocates more space
	for a new subtable
-------------------------------------------------*/

static UINT8 alloc_new_subtable(const struct addrspace_data *space, struct table_data *tabledata, UINT8 previous_value)
{
	/* make sure we don't run out */
	if (tabledata->subtable_count + 1 == SUBTABLE_COUNT)
		fatalerror("error: ran out of memory subtables\n");

	/* allocate more memory if we need to */
	if (tabledata->subtable_count <= tabledata->subtable_alloc)
	{
		tabledata->subtable_alloc += SUBTABLE_ALLOC;
		tabledata->table = realloc(tabledata->table, (1 << LEVEL1_BITS) + (tabledata->subtable_alloc << LEVEL2_BITS));
		if (!tabledata->table)
			fatalerror("error: ran out of memory allocating memory subtable\n");
	}

	/* initialize the table entries */
	memset(&tabledata->table[(1 << LEVEL1_BITS) + (tabledata->subtable_count << LEVEL2_BITS)], previous_value, 1 << LEVEL2_BITS);

	/* return the new index */
	return SUBTABLE_BASE + tabledata->subtable_count++;
}


/*-------------------------------------------------
	allocate_memory - allocate memory for
	sparse CPU address spaces
-------------------------------------------------*/

static int allocate_memory(void)
{
	int cpunum, spacenum;
	ext_entries = 0;

	/* don't do it for drivers that don't have ROM (MESS needs this) */
	if (Machine->gamedrv->rom == 0)
		return 1;

	/* loop over all CPUs and memory spaces */
	for (cpunum = 0; cpunum < cpu_gettotalcpu(); cpunum++)
		for (spacenum = 0; spacenum < ADDRESS_SPACES; spacenum++)
			if (cpudata[cpunum].spacemask & (1 << spacenum))
			{
				struct addrspace_data *space = &cpudata[cpunum].space[spacenum];
				offs_t necessary_size_minus_one = space->mask;
				int i;
					
				/* compute the minimum necessary size for the given address space */
				for (i = 0; i < 32; i++)
					necessary_size_minus_one |= (necessary_size_minus_one >> i);

				/* for non-sparse spaces, we need to have a full memory allocation */
				if (!IS_SPARSE(space->abits))
				{
					/* if this is a program block, make sure the corresponding region */
					if (spacenum == ADDRESS_SPACE_PROGRAM)
					{
						if (memory_region_length(REGION_CPU1 + cpunum) - 1 < necessary_size_minus_one)
							return fatalerror("Memory region REGION_CPU%d too small for address space!\n(currently %X; needs to be %X or bigger)\n", cpunum+1, memory_region_length(REGION_CPU1 + cpunum), necessary_size_minus_one+1);
						register_for_save(cpunum, spacenum, 0, memory_region(REGION_CPU1 + cpunum), necessary_size_minus_one + 1);
					}
					
					/* otherwise, we should just allocate the whole thing */
					else
					{
						if (!allocate_memory_block(cpunum, spacenum, 0, necessary_size_minus_one))
							return 0;
					}
				}
				
				/* for sparse spaces, we find contiguous chunks of memory and allocate them */
				else
				{
					int blocksize = 1 << LEVEL2_BITS;
					int startblock = 0, endblock;
					int lastblock = necessary_size_minus_one / blocksize;
				
					/* if we're in the program space, we already have a head start */
					if (spacenum == ADDRESS_SPACE_PROGRAM)
						startblock = memory_region_length(REGION_CPU1 + cpunum) / blocksize;
					
					/* start looking for contiguous blocks */
					while (1)
					{
						/* first skip over any unmapped or NOP spaces */
						while (startblock <= lastblock &&
								(space->read.table[startblock] == STATIC_NOP || space->read.table[startblock] == STATIC_UNMAP) &&
								(space->write.table[startblock] == STATIC_NOP || space->write.table[startblock] == STATIC_UNMAP))
							startblock++;
						
						/* if we're past the end now, we're done */
						if (startblock > lastblock)
							break;
						
						/* now find the end block */
						endblock = startblock + 1;
						while (endblock <= lastblock &&
								((space->read.table[endblock] != STATIC_NOP && space->read.table[endblock] != STATIC_UNMAP) ||
								 (space->write.table[endblock] != STATIC_NOP && space->write.table[endblock] != STATIC_UNMAP)))
							endblock++;
						
						/* allocate a new contiguous chunk */
						if (!allocate_memory_block(cpunum, spacenum, startblock * blocksize, endblock * blocksize - 1))
							return 0;
						
						/* point to the next one */
						startblock = endblock;
					}
				}
			}

	return 1;
}


/*-------------------------------------------------
	allocate_memory_block - allocate a single
	memory block of data
-------------------------------------------------*/

static int allocate_memory_block(int cpunum, int spacenum, offs_t start, offs_t end)
{
	/* attempt to allocate the memory and error if we fail */
	struct memory_block_t *ext = &ext_memory[ext_entries];
	void *data = malloc(end - start + 1);
	if (!data)
		return fatalerror("Out of memory allocating %d bytes for CPU %d, space %d, range %X-%X\n", end - start + 1, cpunum, spacenum, start, end);
	
	/* clear the memory */
	memset(data, 0, end - start + 1);
	register_for_save(cpunum, spacenum, start, data, end - start + 1);

	/* fill in the tracking block */
	ext->cpunum = cpunum;
	ext->spacenum = spacenum;
	ext->start = start;
	ext->end = end;
	ext->data = data;
	ext_entries++;
	return 1;
}


/*-------------------------------------------------
	register_for_save - register a block of
	memory for save states
-------------------------------------------------*/

static void register_for_save(int cpunum, int spacenum, offs_t start, void *base, size_t numbytes)
{
	char name[256];
	
	sprintf(name, "%d.%08x-%08x", spacenum, start, start + numbytes - 1);
	switch (cpudata[cpunum].space[spacenum].dbits)
	{
		case 8:
			state_save_register_UINT8 ("memory", cpunum, name, base, numbytes);
			break;
		case 16:
			state_save_register_UINT16("memory", cpunum, name, base, numbytes/2);
			break;
		case 32:
			state_save_register_UINT32("memory", cpunum, name, base, numbytes/4);
			break;
		case 64:
//TODO		state_save_register_UINT64("memory", cpunum, name, base, numbytes/8);
			break;
	}
}


/*-------------------------------------------------
	find_memory - find all the requested pointers
	into the final allocated memory
-------------------------------------------------*/

static int find_memory(void)
{
	int cpunum, spacenum, banknum;

	/* loop over CPUs and address spaces */
	for (cpunum = 0; cpunum < cpu_gettotalcpu(); cpunum++)
		for (spacenum = 0; spacenum < ADDRESS_SPACES; spacenum++)
			if (cpudata[cpunum].spacemask & (1 << spacenum))
			{
				struct addrspace_data *space = &cpudata[cpunum].space[spacenum];
				const struct memory_write8_map_t *mwa;
				
				/* first find the end and check for address bits */
				for (mwa = Machine->drv->cpu[cpunum].memory_map[spacenum].write; mwa && !IS_MEMORY_END(mwa); mwa++)
					if (!IS_MEMORY_MARKER(mwa))
					{
						if (mwa->base) *mwa->base = memory_find_base(cpunum, spacenum, SPACE_SHIFT(space, mwa->start));
						if (mwa->size) *mwa->size = SPACE_SHIFT(space, mwa->end - mwa->start + 1);
					}
			}

	/* once this is done, find the starting bases for the banks */
	for (banknum = 1; banknum <= MAX_BANKS; banknum++)
		if (bankdata[banknum].used)
			cpu_bankbase[banknum] = memory_find_base(bankdata[banknum].cpunum, bankdata[banknum].spacenum, bankdata[banknum].base);

	return 1;
}


/*-------------------------------------------------
	memory_find_base - return a pointer to the
	base of RAM associated with the given CPU
	and offset
-------------------------------------------------*/

static void *memory_find_base(int cpunum, int spacenum, offs_t offset)
{
	struct memory_block_t *ext;
	int extnum;

	/* look in external memory first */
	ext = ext_memory;
	for (extnum = 0; extnum < ext_entries; extnum++)
	{
		if (ext->cpunum == cpunum && ext->spacenum == spacenum && ext->start <= offset && ext->end >= offset)
			return (void *)((UINT8 *)ext->data + (offset - ext->start));
		ext++;
	}
	return (UINT8 *)cpudata[cpunum].rambase + offset;
}


/*-------------------------------------------------
	PERFORM_LOOKUP - common lookup procedure
-------------------------------------------------*/

#define PERFORM_LOOKUP(lookup,space,extraand)											\
	/* perform lookup */																\
	address &= space.addrmask & extraand;												\
	entry = space.lookup[LEVEL1_INDEX(address)];										\
	if (entry >= SUBTABLE_BASE)															\
		entry = space.lookup[LEVEL2_INDEX(entry,address)];								\


/*-------------------------------------------------
	READBYTE - generic byte-sized read handler
-------------------------------------------------*/

#define READBYTE8(name,space)															\
data8_t name(offs_t address)															\
{																						\
	UINT8 entry;																		\
	MEMREADSTART();																		\
	PERFORM_LOOKUP(readlookup,space,~0);												\
																						\
	/* handle banks inline */															\
	address -= space.readhandlers[entry].offset;										\
	if (entry <= STATIC_RAM)															\
		MEMREADEND(cpu_bankbase[entry][address]);										\
																						\
	/* fall back to the handler */														\
	else																				\
	{																					\
		read8_handler handler = (read8_handler)space.readhandlers[entry].handler;		\
		MEMREADEND((*handler)(address));												\
	}																					\
	return 0;																			\
}																						\

#define READBYTE(name,space,xormacro,handlertype,ignorebits,shiftbytes)					\
data8_t name(offs_t address)															\
{																						\
	UINT8 entry;																		\
	MEMREADSTART();																		\
	PERFORM_LOOKUP(readlookup,space,~0);												\
																						\
	/* handle banks inline */															\
	address -= space.readhandlers[entry].offset;										\
	if (entry <= STATIC_RAM)															\
		MEMREADEND(cpu_bankbase[entry][xormacro(address)]);								\
																						\
	/* fall back to the handler */														\
	else																				\
	{																					\
		int shift = 8 * (shiftbytes);													\
		handlertype handler = (handlertype)space.readhandlers[entry].handler;			\
		MEMREADEND((*handler)(address >> (ignorebits), ~(0xff << shift)) >> shift);		\
	}																					\
	return 0;																			\
}																						\

#define READBYTE16BE(name,space)	READBYTE(name,space,BYTE_XOR_BE, read16_handler,1,~address & 1)
#define READBYTE16LE(name,space)	READBYTE(name,space,BYTE_XOR_LE, read16_handler,1, address & 1)
#define READBYTE32BE(name,space)	READBYTE(name,space,BYTE4_XOR_BE,read32_handler,2,~address & 3)
#define READBYTE32LE(name,space)	READBYTE(name,space,BYTE4_XOR_LE,read32_handler,2, address & 3)
#define READBYTE64BE(name,space)	READBYTE(name,space,BYTE8_XOR_BE,read64_handler,3,~address & 7)
#define READBYTE64LE(name,space)	READBYTE(name,space,BYTE8_XOR_LE,read64_handler,3, address & 7)


/*-------------------------------------------------
	READWORD - generic word-sized read handler
	(16-bit, 32-bit and 64-bit aligned only!)
-------------------------------------------------*/

#define READWORD16(name,space)															\
data16_t name(offs_t address)															\
{																						\
	UINT8 entry;																		\
	MEMREADSTART();																		\
	PERFORM_LOOKUP(readlookup,space,~1);												\
																						\
	/* handle banks inline */															\
	address -= space.readhandlers[entry].offset;										\
	if (entry <= STATIC_RAM)															\
		MEMREADEND(*(data16_t *)&cpu_bankbase[entry][address]);							\
																						\
	/* fall back to the handler */														\
	else																				\
	{																					\
		read16_handler handler = (read16_handler)space.readhandlers[entry].handler;		\
		MEMREADEND((*handler)(address >> 1,0));										 	\
	}																					\
	return 0;																			\
}																						\

#define READWORD(name,space,xormacro,handlertype,ignorebits,shiftbytes)					\
data16_t name(offs_t address)															\
{																						\
	UINT8 entry;																		\
	MEMREADSTART();																		\
	PERFORM_LOOKUP(readlookup,space,~1);												\
																						\
	/* handle banks inline */															\
	address -= space.readhandlers[entry].offset;										\
	if (entry <= STATIC_RAM)															\
		MEMREADEND(*(data16_t *)&cpu_bankbase[entry][xormacro(address)]);				\
																						\
	/* fall back to the handler */														\
	else																				\
	{																					\
		int shift = 8 * (shiftbytes);													\
		read32_handler handler = (read32_handler)space.readhandlers[entry].handler;		\
		MEMREADEND((*handler)(address >> (ignorebits), ~(0xffff << shift)) >> shift);	\
	}																					\
	return 0;																			\
}																						\

#define READWORD32BE(name,space)	READWORD(name,space,WORD_XOR_BE, read32_handler,2,~address & 2)
#define READWORD32LE(name,space)	READWORD(name,space,WORD_XOR_LE, read32_handler,2, address & 2)
#define READWORD64BE(name,space)	READWORD(name,space,WORD2_XOR_BE,read64_handler,3,~address & 6)
#define READWORD64LE(name,space)	READWORD(name,space,WORD2_XOR_LE,read64_handler,3, address & 6)


/*-------------------------------------------------
	READDWORD - generic dword-sized read handler
	(32-bit and 64-bit aligned only!)
-------------------------------------------------*/

#define READDWORD32(name,space)															\
data32_t name(offs_t address)															\
{																						\
	UINT8 entry;																		\
	MEMREADSTART();																		\
	PERFORM_LOOKUP(readlookup,space,~3);												\
																						\
	/* handle banks inline */															\
	address -= space.readhandlers[entry].offset;										\
	if (entry <= STATIC_RAM)															\
		MEMREADEND(*(data32_t *)&cpu_bankbase[entry][address]);							\
																						\
	/* fall back to the handler */														\
	else																				\
	{																					\
		read32_handler handler = (read32_handler)space.readhandlers[entry].handler;		\
		MEMREADEND((*handler)(address >> 2,0));										 	\
	}																					\
	return 0;																			\
}																						\

#define READDWORD(name,space,xormacro,handlertype,ignorebits,shiftbytes)				\
data32_t name(offs_t address)															\
{																						\
	UINT8 entry;																		\
	MEMREADSTART();																		\
	PERFORM_LOOKUP(readlookup,space,~3);												\
																						\
	/* handle banks inline */															\
	address -= space.readhandlers[entry].offset;										\
	if (entry <= STATIC_RAM)															\
		MEMREADEND(*(data32_t *)&cpu_bankbase[entry][xormacro(address)]);				\
																						\
	/* fall back to the handler */														\
	else																				\
	{																					\
		int shift = 8 * (shiftbytes);													\
		handlertype handler = (handlertype)space.readhandlers[entry].handler;			\
		MEMREADEND((*handler)(address >> (ignorebits), ~(0xffffffff << shift)) >> shift);\
	}																					\
	return 0;																			\
}																						\

#define READDWORD64BE(name,space)	READDWORD(name,space,DWORD_XOR_BE,read64_handler,3,~address & 4)
#define READDWORD64LE(name,space)	READDWORD(name,space,DWORD_XOR_LE,read64_handler,3, address & 4)


/*-------------------------------------------------
	READQWORD - generic qword-sized read handler
	(64-bit aligned only!)
-------------------------------------------------*/

#define READQWORD64(name,space)															\
data64_t name(offs_t address)															\
{																						\
	UINT8 entry;																		\
	MEMREADSTART();																		\
	PERFORM_LOOKUP(readlookup,space,~7);												\
																						\
	/* handle banks inline */															\
	address -= space.readhandlers[entry].offset;										\
	if (entry <= STATIC_RAM)															\
		MEMREADEND(*(data64_t *)&cpu_bankbase[entry][address]);							\
																						\
	/* fall back to the handler */														\
	else																				\
	{																					\
		read64_handler handler = (read64_handler)space.readhandlers[entry].handler;		\
		MEMREADEND((*handler)(address >> 3,0));										 	\
	}																					\
	return 0;																			\
}																						\


/*-------------------------------------------------
	WRITEBYTE - generic byte-sized write handler
-------------------------------------------------*/

#define WRITEBYTE8(name,space)															\
void name(offs_t address, data8_t data)													\
{																						\
	UINT8 entry;																		\
	MEMWRITESTART();																	\
	PERFORM_LOOKUP(writelookup,space,~0);												\
																						\
	/* handle banks inline */															\
	address -= space.writehandlers[entry].offset;										\
	if (entry <= STATIC_RAM)															\
		MEMWRITEEND(cpu_bankbase[entry][address] = data);								\
																						\
	/* fall back to the handler */														\
	else																				\
	{																					\
		write8_handler handler = (write8_handler)space.writehandlers[entry].handler;	\
		MEMWRITEEND((*handler)(address, data));											\
	}																					\
}																						\

#define WRITEBYTE(name,space,xormacro,handlertype,ignorebits,shiftbytes)				\
void name(offs_t address, data8_t data)													\
{																						\
	UINT8 entry;																		\
	MEMWRITESTART();																	\
	PERFORM_LOOKUP(writelookup,space,~0);												\
																						\
	/* handle banks inline */															\
	address -= space.writehandlers[entry].offset;										\
	if (entry <= STATIC_RAM)															\
		MEMWRITEEND(cpu_bankbase[entry][xormacro(address)] = data);						\
																						\
	/* fall back to the handler */														\
	else																				\
	{																					\
		int shift = 8 * (shiftbytes);													\
		handlertype handler = (handlertype)space.writehandlers[entry].handler;			\
		MEMWRITEEND((*handler)(address >> (ignorebits), data << shift, ~(0xff << shift)));\
	}																					\
}																						\

#define WRITEBYTE16BE(name,space)	WRITEBYTE(name,space,BYTE_XOR_BE, write16_handler,1,~address & 1)
#define WRITEBYTE16LE(name,space)	WRITEBYTE(name,space,BYTE_XOR_LE, write16_handler,1, address & 1)
#define WRITEBYTE32BE(name,space)	WRITEBYTE(name,space,BYTE4_XOR_BE,write32_handler,2,~address & 3)
#define WRITEBYTE32LE(name,space)	WRITEBYTE(name,space,BYTE4_XOR_LE,write32_handler,2, address & 3)
#define WRITEBYTE64BE(name,space)	WRITEBYTE(name,space,BYTE8_XOR_BE,write64_handler,3,~address & 7)
#define WRITEBYTE64LE(name,space)	WRITEBYTE(name,space,BYTE8_XOR_LE,write64_handler,3, address & 7)


/*-------------------------------------------------
	WRITEWORD - generic word-sized write handler
	(16-bit, 32-bit and 64-bit aligned only!)
-------------------------------------------------*/

#define WRITEWORD16(name,space)															\
void name(offs_t address, data16_t data)												\
{																						\
	UINT8 entry;																		\
	MEMWRITESTART();																	\
	PERFORM_LOOKUP(writelookup,space,~1);												\
																						\
	/* handle banks inline */															\
	address -= space.writehandlers[entry].offset;										\
	if (entry <= STATIC_RAM)															\
		MEMWRITEEND(*(data16_t *)&cpu_bankbase[entry][address] = data);					\
																						\
	/* fall back to the handler */														\
	else																				\
	{																					\
		write16_handler handler = (write16_handler)space.writehandlers[entry].handler;	\
		MEMWRITEEND((*handler)(address >> 1, data, 0));								 	\
	}																					\
}																						\

#define WRITEWORD(name,space,xormacro,handlertype,ignorebits,shiftbytes)				\
void name(offs_t address, data16_t data)												\
{																						\
	UINT8 entry;																		\
	MEMWRITESTART();																	\
	PERFORM_LOOKUP(writelookup,space,~1);												\
																						\
	/* handle banks inline */															\
	address -= space.writehandlers[entry].offset;										\
	if (entry <= STATIC_RAM)															\
		MEMWRITEEND(*(data16_t *)&cpu_bankbase[entry][xormacro(address)] = data);		\
																						\
	/* fall back to the handler */														\
	else																				\
	{																					\
		int shift = 8 * (shiftbytes);													\
		handlertype handler = (handlertype)space.writehandlers[entry].handler;			\
		MEMWRITEEND((*handler)(address >> (ignorebits), data << shift, ~(0xffff << shift)));\
	}																					\
}																						\

#define WRITEWORD32BE(name,space)	WRITEWORD(name,space,WORD_XOR_BE, write32_handler,2,~address & 2)
#define WRITEWORD32LE(name,space)	WRITEWORD(name,space,WORD_XOR_LE, write32_handler,2, address & 2)
#define WRITEWORD64BE(name,space)	WRITEWORD(name,space,WORD2_XOR_BE,write64_handler,3,~address & 6)
#define WRITEWORD64LE(name,space)	WRITEWORD(name,space,WORD2_XOR_LE,write64_handler,3, address & 6)


/*-------------------------------------------------
	WRITEDWORD - dword-sized write handler
	(32-bit and 64-bit aligned only!)
-------------------------------------------------*/

#define WRITEDWORD32(name,space)														\
void name(offs_t address, data32_t data)												\
{																						\
	UINT8 entry;																		\
	MEMWRITESTART();																	\
	PERFORM_LOOKUP(writelookup,space,~3);												\
																						\
	/* handle banks inline */															\
	address -= space.writehandlers[entry].offset;										\
	if (entry <= STATIC_RAM)															\
		MEMWRITEEND(*(data32_t *)&cpu_bankbase[entry][address] = data);					\
																						\
	/* fall back to the handler */														\
	else																				\
	{																					\
		write32_handler handler = (write32_handler)space.writehandlers[entry].handler;	\
		MEMWRITEEND((*handler)(address >> 2, data, 0));								 	\
	}																					\
}																						\

#define WRITEDWORD(name,space,xormacro,handlertype,ignorebits,shiftbytes)				\
void name(offs_t address, data32_t data)												\
{																						\
	UINT8 entry;																		\
	MEMWRITESTART();																	\
	PERFORM_LOOKUP(writelookup,space,~3);												\
																						\
	/* handle banks inline */															\
	address -= space.writehandlers[entry].offset;										\
	if (entry <= STATIC_RAM)															\
		MEMWRITEEND(*(data32_t *)&cpu_bankbase[entry][xormacro(address)] = data);		\
																						\
	/* fall back to the handler */														\
	else																				\
	{																					\
		int shift = 8 * (shiftbytes);													\
		handlertype handler = (handlertype)space.writehandlers[entry].handler;			\
		MEMWRITEEND((*handler)(address >> (ignorebits), data << shift, ~(0xffffffff << shift)));\
	}																					\
}																						\

#define WRITEDWORD64BE(name,space)	WRITEDWORD(name,space,DWORD_XOR_BE,write64_handler,3,~address & 4)
#define WRITEDWORD64LE(name,space)	WRITEDWORD(name,space,DWORD_XOR_LE,write64_handler,3, address & 4)


/*-------------------------------------------------
	WRITEQWORD - qword-sized write handler
	(64-bit aligned only!)
-------------------------------------------------*/

#define WRITEQWORD64(name,space)														\
void name(offs_t address, data64_t data)												\
{																						\
	UINT8 entry;																		\
	MEMWRITESTART();																	\
	PERFORM_LOOKUP(writelookup,space,~7);												\
																						\
	/* handle banks inline */															\
	address -= space.writehandlers[entry].offset;										\
	if (entry <= STATIC_RAM)															\
		MEMWRITEEND(*(data64_t *)&cpu_bankbase[entry][address] = data);					\
																						\
	/* fall back to the handler */														\
	else																				\
	{																					\
		write64_handler handler = (write64_handler)space.writehandlers[entry].handler;	\
		MEMWRITEEND((*handler)(address >> 3, data, 0));								 	\
	}																					\
}																						\


/*-------------------------------------------------
	Program memory handlers
-------------------------------------------------*/

     READBYTE8(program_read_byte_8,      address_space[ADDRESS_SPACE_PROGRAM])
    WRITEBYTE8(program_write_byte_8,     address_space[ADDRESS_SPACE_PROGRAM])

  READBYTE16BE(program_read_byte_16be,   address_space[ADDRESS_SPACE_PROGRAM])
    READWORD16(program_read_word_16be,   address_space[ADDRESS_SPACE_PROGRAM])
 WRITEBYTE16BE(program_write_byte_16be,  address_space[ADDRESS_SPACE_PROGRAM])
   WRITEWORD16(program_write_word_16be,  address_space[ADDRESS_SPACE_PROGRAM])
   
  READBYTE16LE(program_read_byte_16le,   address_space[ADDRESS_SPACE_PROGRAM])
    READWORD16(program_read_word_16le,   address_space[ADDRESS_SPACE_PROGRAM])
 WRITEBYTE16LE(program_write_byte_16le,  address_space[ADDRESS_SPACE_PROGRAM])
   WRITEWORD16(program_write_word_16le,  address_space[ADDRESS_SPACE_PROGRAM])

  READBYTE32BE(program_read_byte_32be,   address_space[ADDRESS_SPACE_PROGRAM])
  READWORD32BE(program_read_word_32be,   address_space[ADDRESS_SPACE_PROGRAM])
   READDWORD32(program_read_dword_32be,  address_space[ADDRESS_SPACE_PROGRAM])
 WRITEBYTE32BE(program_write_byte_32be,  address_space[ADDRESS_SPACE_PROGRAM])
 WRITEWORD32BE(program_write_word_32be,  address_space[ADDRESS_SPACE_PROGRAM])
  WRITEDWORD32(program_write_dword_32be, address_space[ADDRESS_SPACE_PROGRAM])
   
  READBYTE32LE(program_read_byte_32le,   address_space[ADDRESS_SPACE_PROGRAM])
  READWORD32LE(program_read_word_32le,   address_space[ADDRESS_SPACE_PROGRAM])
   READDWORD32(program_read_dword_32le,  address_space[ADDRESS_SPACE_PROGRAM])
 WRITEBYTE32LE(program_write_byte_32le,  address_space[ADDRESS_SPACE_PROGRAM])
 WRITEWORD32LE(program_write_word_32le,  address_space[ADDRESS_SPACE_PROGRAM])
  WRITEDWORD32(program_write_dword_32le, address_space[ADDRESS_SPACE_PROGRAM])
   
  READBYTE64BE(program_read_byte_64be,   address_space[ADDRESS_SPACE_PROGRAM])
  READWORD64BE(program_read_word_64be,   address_space[ADDRESS_SPACE_PROGRAM])
 READDWORD64BE(program_read_dword_64be,  address_space[ADDRESS_SPACE_PROGRAM])
   READQWORD64(program_read_qword_64be,  address_space[ADDRESS_SPACE_PROGRAM])
 WRITEBYTE64BE(program_write_byte_64be,  address_space[ADDRESS_SPACE_PROGRAM])
 WRITEWORD64BE(program_write_word_64be,  address_space[ADDRESS_SPACE_PROGRAM])
WRITEDWORD64BE(program_write_dword_64be, address_space[ADDRESS_SPACE_PROGRAM])
  WRITEQWORD64(program_write_qword_64be, address_space[ADDRESS_SPACE_PROGRAM])
   
  READBYTE64LE(program_read_byte_64le,   address_space[ADDRESS_SPACE_PROGRAM])
  READWORD64LE(program_read_word_64le,   address_space[ADDRESS_SPACE_PROGRAM])
 READDWORD64LE(program_read_dword_64le,  address_space[ADDRESS_SPACE_PROGRAM])
   READQWORD64(program_read_qword_64le,  address_space[ADDRESS_SPACE_PROGRAM])
 WRITEBYTE64LE(program_write_byte_64le,  address_space[ADDRESS_SPACE_PROGRAM])
 WRITEWORD64LE(program_write_word_64le,  address_space[ADDRESS_SPACE_PROGRAM])
WRITEDWORD64LE(program_write_dword_64le, address_space[ADDRESS_SPACE_PROGRAM])
  WRITEQWORD64(program_write_qword_64le, address_space[ADDRESS_SPACE_PROGRAM])


/*-------------------------------------------------
	Data memory handlers
-------------------------------------------------*/

     READBYTE8(data_read_byte_8,      address_space[ADDRESS_SPACE_DATA])
    WRITEBYTE8(data_write_byte_8,     address_space[ADDRESS_SPACE_DATA])

  READBYTE16BE(data_read_byte_16be,   address_space[ADDRESS_SPACE_DATA])
    READWORD16(data_read_word_16be,   address_space[ADDRESS_SPACE_DATA])
 WRITEBYTE16BE(data_write_byte_16be,  address_space[ADDRESS_SPACE_DATA])
   WRITEWORD16(data_write_word_16be,  address_space[ADDRESS_SPACE_DATA])
   
  READBYTE16LE(data_read_byte_16le,   address_space[ADDRESS_SPACE_DATA])
    READWORD16(data_read_word_16le,   address_space[ADDRESS_SPACE_DATA])
 WRITEBYTE16LE(data_write_byte_16le,  address_space[ADDRESS_SPACE_DATA])
   WRITEWORD16(data_write_word_16le,  address_space[ADDRESS_SPACE_DATA])

  READBYTE32BE(data_read_byte_32be,   address_space[ADDRESS_SPACE_DATA])
  READWORD32BE(data_read_word_32be,   address_space[ADDRESS_SPACE_DATA])
   READDWORD32(data_read_dword_32be,  address_space[ADDRESS_SPACE_DATA])
 WRITEBYTE32BE(data_write_byte_32be,  address_space[ADDRESS_SPACE_DATA])
 WRITEWORD32BE(data_write_word_32be,  address_space[ADDRESS_SPACE_DATA])
  WRITEDWORD32(data_write_dword_32be, address_space[ADDRESS_SPACE_DATA])
   
  READBYTE32LE(data_read_byte_32le,   address_space[ADDRESS_SPACE_DATA])
  READWORD32LE(data_read_word_32le,   address_space[ADDRESS_SPACE_DATA])
   READDWORD32(data_read_dword_32le,  address_space[ADDRESS_SPACE_DATA])
 WRITEBYTE32LE(data_write_byte_32le,  address_space[ADDRESS_SPACE_DATA])
 WRITEWORD32LE(data_write_word_32le,  address_space[ADDRESS_SPACE_DATA])
  WRITEDWORD32(data_write_dword_32le, address_space[ADDRESS_SPACE_DATA])
   
  READBYTE64BE(data_read_byte_64be,   address_space[ADDRESS_SPACE_DATA])
  READWORD64BE(data_read_word_64be,   address_space[ADDRESS_SPACE_DATA])
 READDWORD64BE(data_read_dword_64be,  address_space[ADDRESS_SPACE_DATA])
   READQWORD64(data_read_qword_64be,  address_space[ADDRESS_SPACE_DATA])
 WRITEBYTE64BE(data_write_byte_64be,  address_space[ADDRESS_SPACE_DATA])
 WRITEWORD64BE(data_write_word_64be,  address_space[ADDRESS_SPACE_DATA])
WRITEDWORD64BE(data_write_dword_64be, address_space[ADDRESS_SPACE_DATA])
  WRITEQWORD64(data_write_qword_64be, address_space[ADDRESS_SPACE_DATA])
   
  READBYTE64LE(data_read_byte_64le,   address_space[ADDRESS_SPACE_DATA])
  READWORD64LE(data_read_word_64le,   address_space[ADDRESS_SPACE_DATA])
 READDWORD64LE(data_read_dword_64le,  address_space[ADDRESS_SPACE_DATA])
   READQWORD64(data_read_qword_64le,  address_space[ADDRESS_SPACE_DATA])
 WRITEBYTE64LE(data_write_byte_64le,  address_space[ADDRESS_SPACE_DATA])
 WRITEWORD64LE(data_write_word_64le,  address_space[ADDRESS_SPACE_DATA])
WRITEDWORD64LE(data_write_dword_64le, address_space[ADDRESS_SPACE_DATA])
  WRITEQWORD64(data_write_qword_64le, address_space[ADDRESS_SPACE_DATA])


/*-------------------------------------------------
	I/O memory handlers
-------------------------------------------------*/

     READBYTE8(io_read_byte_8,      address_space[ADDRESS_SPACE_IO])
    WRITEBYTE8(io_write_byte_8,     address_space[ADDRESS_SPACE_IO])

  READBYTE16BE(io_read_byte_16be,   address_space[ADDRESS_SPACE_IO])
    READWORD16(io_read_word_16be,   address_space[ADDRESS_SPACE_IO])
 WRITEBYTE16BE(io_write_byte_16be,  address_space[ADDRESS_SPACE_IO])
   WRITEWORD16(io_write_word_16be,  address_space[ADDRESS_SPACE_IO])
   
  READBYTE16LE(io_read_byte_16le,   address_space[ADDRESS_SPACE_IO])
    READWORD16(io_read_word_16le,   address_space[ADDRESS_SPACE_IO])
 WRITEBYTE16LE(io_write_byte_16le,  address_space[ADDRESS_SPACE_IO])
   WRITEWORD16(io_write_word_16le,  address_space[ADDRESS_SPACE_IO])

  READBYTE32BE(io_read_byte_32be,   address_space[ADDRESS_SPACE_IO])
  READWORD32BE(io_read_word_32be,   address_space[ADDRESS_SPACE_IO])
   READDWORD32(io_read_dword_32be,  address_space[ADDRESS_SPACE_IO])
 WRITEBYTE32BE(io_write_byte_32be,  address_space[ADDRESS_SPACE_IO])
 WRITEWORD32BE(io_write_word_32be,  address_space[ADDRESS_SPACE_IO])
  WRITEDWORD32(io_write_dword_32be, address_space[ADDRESS_SPACE_IO])
   
  READBYTE32LE(io_read_byte_32le,   address_space[ADDRESS_SPACE_IO])
  READWORD32LE(io_read_word_32le,   address_space[ADDRESS_SPACE_IO])
   READDWORD32(io_read_dword_32le,  address_space[ADDRESS_SPACE_IO])
 WRITEBYTE32LE(io_write_byte_32le,  address_space[ADDRESS_SPACE_IO])
 WRITEWORD32LE(io_write_word_32le,  address_space[ADDRESS_SPACE_IO])
  WRITEDWORD32(io_write_dword_32le, address_space[ADDRESS_SPACE_IO])
   
  READBYTE64BE(io_read_byte_64be,   address_space[ADDRESS_SPACE_IO])
  READWORD64BE(io_read_word_64be,   address_space[ADDRESS_SPACE_IO])
 READDWORD64BE(io_read_dword_64be,  address_space[ADDRESS_SPACE_IO])
   READQWORD64(io_read_qword_64be,  address_space[ADDRESS_SPACE_IO])
 WRITEBYTE64BE(io_write_byte_64be,  address_space[ADDRESS_SPACE_IO])
 WRITEWORD64BE(io_write_word_64be,  address_space[ADDRESS_SPACE_IO])
WRITEDWORD64BE(io_write_dword_64be, address_space[ADDRESS_SPACE_IO])
  WRITEQWORD64(io_write_qword_64be, address_space[ADDRESS_SPACE_IO])
   
  READBYTE64LE(io_read_byte_64le,   address_space[ADDRESS_SPACE_IO])
  READWORD64LE(io_read_word_64le,   address_space[ADDRESS_SPACE_IO])
 READDWORD64LE(io_read_dword_64le,  address_space[ADDRESS_SPACE_IO])
   READQWORD64(io_read_qword_64le,  address_space[ADDRESS_SPACE_IO])
 WRITEBYTE64LE(io_write_byte_64le,  address_space[ADDRESS_SPACE_IO])
 WRITEWORD64LE(io_write_word_64le,  address_space[ADDRESS_SPACE_IO])
WRITEDWORD64LE(io_write_dword_64le, address_space[ADDRESS_SPACE_IO])
  WRITEQWORD64(io_write_qword_64le, address_space[ADDRESS_SPACE_IO])


/*-------------------------------------------------
	safe opcode reading
-------------------------------------------------*/

data8_t cpu_readop_safe(offs_t offset)
{
	activecpu_set_opbase(offset);
	return cpu_readop_unsafe(offset);
}

data16_t cpu_readop16_safe(offs_t offset)
{
	activecpu_set_opbase(offset);
	return cpu_readop16_unsafe(offset);
}

data32_t cpu_readop32_safe(offs_t offset)
{
	activecpu_set_opbase(offset);
	return cpu_readop32_unsafe(offset);
}

data64_t cpu_readop64_safe(offs_t offset)
{
	activecpu_set_opbase(offset);
	return cpu_readop64_unsafe(offset);
}

data8_t cpu_readop_arg_safe(offs_t offset)
{
	activecpu_set_opbase(offset);
	return cpu_readop_arg_unsafe(offset);
}

data16_t cpu_readop_arg16_safe(offs_t offset)
{
	activecpu_set_opbase(offset);
	return cpu_readop_arg16_unsafe(offset);
}

data32_t cpu_readop_arg32_safe(offs_t offset)
{
	activecpu_set_opbase(offset);
	return cpu_readop_arg32_unsafe(offset);
}

data64_t cpu_readop_arg64_safe(offs_t offset)
{
	activecpu_set_opbase(offset);
	return cpu_readop_arg64_unsafe(offset);
}


/*-------------------------------------------------
	unmapped memory handlers
-------------------------------------------------*/

static READ_HANDLER( mrh8_unmap_program )
{
	logerror("cpu #%d (PC=%08X): unmapped program memory byte read from %08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM], offset));
	return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM].unmap;
}
static READ16_HANDLER( mrh16_unmap_program )
{
	logerror("cpu #%d (PC=%08X): unmapped program memory word read from %08X & %04X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM], offset*2), mem_mask ^ 0xffff);
	return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM].unmap;
}
static READ32_HANDLER( mrh32_unmap_program )
{
	logerror("cpu #%d (PC=%08X): unmapped program memory dword read from %08X & %08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM], offset*4), mem_mask ^ 0xffffffff);
	return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM].unmap;
}
static READ64_HANDLER( mrh64_unmap_program )
{
	logerror("cpu #%d (PC=%08X): unmapped program memory qword read from %08X & %08X%08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM], offset*8), (int)(mem_mask >> 32) ^ 0xffffffff, (int)(mem_mask & 0xffffffff) ^ 0xffffffff);
	return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM].unmap;
}

static WRITE_HANDLER( mwh8_unmap_program )
{
	logerror("cpu #%d (PC=%08X): unmapped program memory byte write to %08X = %02X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM], offset), data);
}
static WRITE16_HANDLER( mwh16_unmap_program )
{
	logerror("cpu #%d (PC=%08X): unmapped program memory word write to %08X = %04X & %04X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM], offset*2), data, mem_mask ^ 0xffff);
}
static WRITE32_HANDLER( mwh32_unmap_program )
{
	logerror("cpu #%d (PC=%08X): unmapped program memory dword write to %08X = %08X & %08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM], offset*4), data, mem_mask ^ 0xffffffff);
}
static WRITE64_HANDLER( mwh64_unmap_program )
{
	logerror("cpu #%d (PC=%08X): unmapped program memory qword write to %08X = %08X%08X & %08X%08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM], offset*8), (int)(data >> 32), (int)(data & 0xffffffff), (int)(mem_mask >> 32) ^ 0xffffffff, (int)(mem_mask & 0xffffffff) ^ 0xffffffff);
}

static READ_HANDLER( mrh8_unmap_data )
{
	logerror("cpu #%d (PC=%08X): unmapped data memory byte read from %08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA], offset));
	return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA].unmap;
}
static READ16_HANDLER( mrh16_unmap_data )
{
	logerror("cpu #%d (PC=%08X): unmapped data memory word read from %08X & %04X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA], offset*2), mem_mask ^ 0xffff);
	return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA].unmap;
}
static READ32_HANDLER( mrh32_unmap_data )
{
	logerror("cpu #%d (PC=%08X): unmapped data memory dword read from %08X & %08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA], offset*4), mem_mask ^ 0xffffffff);
	return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA].unmap;
}
static READ64_HANDLER( mrh64_unmap_data )
{
	logerror("cpu #%d (PC=%08X): unmapped data memory qword read from %08X & %08X%08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA], offset*8), (int)(mem_mask >> 32) ^ 0xffffffff, (int)(mem_mask & 0xffffffff) ^ 0xffffffff);
	return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA].unmap;
}

static WRITE_HANDLER( mwh8_unmap_data )
{
	logerror("cpu #%d (PC=%08X): unmapped data memory byte write to %08X = %02X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA], offset), data);
}
static WRITE16_HANDLER( mwh16_unmap_data )
{
	logerror("cpu #%d (PC=%08X): unmapped data memory word write to %08X = %04X & %04X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA], offset*2), data, mem_mask ^ 0xffff);
}
static WRITE32_HANDLER( mwh32_unmap_data )
{
	logerror("cpu #%d (PC=%08X): unmapped data memory dword write to %08X = %08X & %08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA], offset*4), data, mem_mask ^ 0xffffffff);
}
static WRITE64_HANDLER( mwh64_unmap_data )
{
	logerror("cpu #%d (PC=%08X): unmapped data memory qword write to %08X = %08X%08X & %08X%08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA], offset*8), (int)(data >> 32), (int)(data & 0xffffffff), (int)(mem_mask >> 32) ^ 0xffffffff, (int)(mem_mask & 0xffffffff) ^ 0xffffffff);
}

static READ_HANDLER( mrh8_unmap_io )
{
	logerror("cpu #%d (PC=%08X): unmapped I/O byte read from %08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO], offset));
	return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO].unmap;
}
static READ16_HANDLER( mrh16_unmap_io )
{
	logerror("cpu #%d (PC=%08X): unmapped I/O word read from %08X & %04X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO], offset*2), mem_mask ^ 0xffff);
	return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO].unmap;
}
static READ32_HANDLER( mrh32_unmap_io )
{
	logerror("cpu #%d (PC=%08X): unmapped I/O dword read from %08X & %08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO], offset*4), mem_mask ^ 0xffffffff);
	return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO].unmap;
}
static READ64_HANDLER( mrh64_unmap_io )
{
	logerror("cpu #%d (PC=%08X): unmapped I/O qword read from %08X & %08X%08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO], offset*8), (int)(mem_mask >> 32) ^ 0xffffffff, (int)(mem_mask & 0xffffffff) ^ 0xffffffff);
	return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO].unmap;
}

static WRITE_HANDLER( mwh8_unmap_io )
{
	logerror("cpu #%d (PC=%08X): unmapped I/O byte write to %08X = %02X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO], offset), data);
}
static WRITE16_HANDLER( mwh16_unmap_io )
{
	logerror("cpu #%d (PC=%08X): unmapped I/O word write to %08X = %04X & %04X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO], offset*2), data, mem_mask ^ 0xffff);
}
static WRITE32_HANDLER( mwh32_unmap_io )
{
	logerror("cpu #%d (PC=%08X): unmapped I/O dword write to %08X = %08X & %08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO], offset*4), data, mem_mask ^ 0xffffffff);
}
static WRITE64_HANDLER( mwh64_unmap_io )
{
	logerror("cpu #%d (PC=%08X): unmapped I/O qword write to %08X = %08X%08X & %08X%08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO], offset*8), (int)(data >> 32), (int)(data & 0xffffffff), (int)(mem_mask >> 32) ^ 0xffffffff, (int)(mem_mask & 0xffffffff) ^ 0xffffffff);
}


/*-------------------------------------------------
	no-op memory handlers
-------------------------------------------------*/

static READ_HANDLER( mrh8_nop_program )    { return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM].unmap; }
static READ16_HANDLER( mrh16_nop_program ) { return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM].unmap; }
static READ32_HANDLER( mrh32_nop_program ) { return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM].unmap; }
static READ64_HANDLER( mrh64_nop_program ) { return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM].unmap; }

static READ_HANDLER( mrh8_nop_data )       { return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA].unmap; }
static READ16_HANDLER( mrh16_nop_data )    { return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA].unmap; }
static READ32_HANDLER( mrh32_nop_data )    { return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA].unmap; }
static READ64_HANDLER( mrh64_nop_data )    { return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_DATA].unmap; }

static READ_HANDLER( mrh8_nop_io )         { return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO].unmap; }
static READ16_HANDLER( mrh16_nop_io )      { return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO].unmap; }
static READ32_HANDLER( mrh32_nop_io )      { return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO].unmap; }
static READ64_HANDLER( mrh64_nop_io )      { return cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_IO].unmap; }

static WRITE_HANDLER( mwh8_nop )           {  }
static WRITE16_HANDLER( mwh16_nop )        {  }
static WRITE32_HANDLER( mwh32_nop )        {  }
static WRITE64_HANDLER( mwh64_nop )        {  }


/*-------------------------------------------------
	other static handlers
-------------------------------------------------*/

static WRITE_HANDLER( mwh8_rom )       { logerror("cpu #%d (PC=%08X): byte write to ROM %08X = %02X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM], offset), data); }
static WRITE16_HANDLER( mwh16_rom )    { logerror("cpu #%d (PC=%08X): word write to %08X = %04X & %04X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM], offset*2), data, mem_mask ^ 0xffff); }
static WRITE32_HANDLER( mwh32_rom )    { logerror("cpu #%d (PC=%08X): dword write to %08X = %08X & %08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM], offset*4), data, mem_mask ^ 0xffffffff); }
static WRITE64_HANDLER( mwh64_rom )    { logerror("cpu #%d (PC=%08X): qword write to %08X = %08X%08X & %08X%08X\n", cpu_getactivecpu(), activecpu_get_pc(), INV_SPACE_SHIFT(&cpudata[cpu_getactivecpu()].space[ADDRESS_SPACE_PROGRAM], offset*8), (int)(data >> 32), (int)(data & 0xffffffff), (int)(mem_mask >> 32) ^ 0xffffffff, (int)(mem_mask & 0xffffffff) ^ 0xffffffff); }

static WRITE_HANDLER( mwh8_ramrom )    { cpu_bankbase[STATIC_RAM][offset] = cpu_bankbase[STATIC_RAM][offset + (OP_ROM - OP_RAM)] = data; }
static WRITE16_HANDLER( mwh16_ramrom ) { COMBINE_DATA(&cpu_bankbase[STATIC_RAM][offset*2]); COMBINE_DATA(&cpu_bankbase[0][offset*2 + (OP_ROM - OP_RAM)]); }
static WRITE32_HANDLER( mwh32_ramrom ) { COMBINE_DATA(&cpu_bankbase[STATIC_RAM][offset*4]); COMBINE_DATA(&cpu_bankbase[0][offset*4 + (OP_ROM - OP_RAM)]); }
static WRITE64_HANDLER( mwh64_ramrom ) { COMBINE_DATA(&cpu_bankbase[STATIC_RAM][offset*8]); COMBINE_DATA(&cpu_bankbase[0][offset*8 + (OP_ROM - OP_RAM)]); }


/*-------------------------------------------------
	get_static_handler - returns points to static
	memory handlers
-------------------------------------------------*/

static void *get_static_handler(int databits, int readorwrite, int spacenum, int which)
{
	static const struct
	{
		UINT8		databits;
		UINT8		handlernum;
		UINT8		spacenum;
		void *		read;
		void *		write;
	} static_handler_list[] =
	{
		{  8, STATIC_UNMAP,  ADDRESS_SPACE_PROGRAM, (void *)mrh8_unmap_program, (void *)mwh8_unmap_program },
		{  8, STATIC_UNMAP,  ADDRESS_SPACE_DATA,    (void *)mrh8_unmap_data,    (void *)mwh8_unmap_data },
		{  8, STATIC_UNMAP,  ADDRESS_SPACE_IO,      (void *)mrh8_unmap_io,      (void *)mwh8_unmap_io },
		{  8, STATIC_NOP,    ADDRESS_SPACE_PROGRAM, (void *)mrh8_nop_program,   (void *)mwh8_nop },
		{  8, STATIC_NOP,    ADDRESS_SPACE_DATA,    (void *)mrh8_nop_data,      (void *)mwh8_nop },
		{  8, STATIC_NOP,    ADDRESS_SPACE_IO,      (void *)mrh8_nop_io,        (void *)mwh8_nop },
		{  8, STATIC_ROM,    0xff,                  NULL,                       (void *)mwh8_rom },
		{  8, STATIC_RAMROM, 0xff,                  NULL,                       (void *)mwh8_ramrom },
		
		{ 16, STATIC_UNMAP,  ADDRESS_SPACE_PROGRAM, (void *)mrh16_unmap_program,(void *)mwh16_unmap_program },
		{ 16, STATIC_UNMAP,  ADDRESS_SPACE_DATA,    (void *)mrh16_unmap_data,   (void *)mwh16_unmap_data },
		{ 16, STATIC_UNMAP,  ADDRESS_SPACE_IO,      (void *)mrh16_unmap_io,     (void *)mwh16_unmap_io },
		{ 16, STATIC_NOP,    ADDRESS_SPACE_PROGRAM, (void *)mrh16_nop_program,  (void *)mwh16_nop },
		{ 16, STATIC_NOP,    ADDRESS_SPACE_DATA,    (void *)mrh16_nop_data,     (void *)mwh16_nop },
		{ 16, STATIC_NOP,    ADDRESS_SPACE_IO,      (void *)mrh16_nop_io,       (void *)mwh16_nop },
		{ 16, STATIC_ROM,    0xff,                  NULL,                       (void *)mwh16_rom },
		{ 16, STATIC_RAMROM, 0xff,                  NULL,                       (void *)mwh16_ramrom },
		
		{ 32, STATIC_UNMAP,  ADDRESS_SPACE_PROGRAM, (void *)mrh32_unmap_program,(void *)mwh32_unmap_program },
		{ 32, STATIC_UNMAP,  ADDRESS_SPACE_DATA,    (void *)mrh32_unmap_data,   (void *)mwh32_unmap_data },
		{ 32, STATIC_UNMAP,  ADDRESS_SPACE_IO,      (void *)mrh32_unmap_io,     (void *)mwh32_unmap_io },
		{ 32, STATIC_NOP,    ADDRESS_SPACE_PROGRAM, (void *)mrh32_nop_program,  (void *)mwh32_nop },
		{ 32, STATIC_NOP,    ADDRESS_SPACE_DATA,    (void *)mrh32_nop_data,     (void *)mwh32_nop },
		{ 32, STATIC_NOP,    ADDRESS_SPACE_IO,      (void *)mrh32_nop_io,       (void *)mwh32_nop },
		{ 32, STATIC_ROM,    0xff,                  NULL,                       (void *)mwh32_rom },
		{ 32, STATIC_RAMROM, 0xff,                  NULL,                       (void *)mwh32_ramrom },
		
		{ 64, STATIC_UNMAP,  ADDRESS_SPACE_PROGRAM, (void *)mrh64_unmap_program,(void *)mwh64_unmap_program },
		{ 64, STATIC_UNMAP,  ADDRESS_SPACE_DATA,    (void *)mrh64_unmap_data,   (void *)mwh64_unmap_data },
		{ 64, STATIC_UNMAP,  ADDRESS_SPACE_IO,      (void *)mrh64_unmap_io,     (void *)mwh64_unmap_io },
		{ 64, STATIC_NOP,    ADDRESS_SPACE_PROGRAM, (void *)mrh64_nop_program,  (void *)mwh64_nop },
		{ 64, STATIC_NOP,    ADDRESS_SPACE_DATA,    (void *)mrh64_nop_data,     (void *)mwh64_nop },
		{ 64, STATIC_NOP,    ADDRESS_SPACE_IO,      (void *)mrh64_nop_io,       (void *)mwh64_nop },
		{ 64, STATIC_ROM,    0xff,                  NULL,                       (void *)mwh64_rom },
		{ 64, STATIC_RAMROM, 0xff,                  NULL,                       (void *)mwh64_ramrom }
	};
	int tablenum;
	
	for (tablenum = 0; tablenum < sizeof(static_handler_list) / sizeof(static_handler_list[0]); tablenum++)
		if (static_handler_list[tablenum].databits == databits && static_handler_list[tablenum].handlernum == which)
			if (static_handler_list[tablenum].spacenum == 0xff || static_handler_list[tablenum].spacenum == spacenum)
				return readorwrite ? static_handler_list[tablenum].write : static_handler_list[tablenum].read;
	
	return NULL;
}


/*-------------------------------------------------
	debugging
-------------------------------------------------*/

#if MEM_DUMP

static void dump_map(FILE *file, const struct addrspace_data *space, const struct table_data *table)
{
	static const char *strings[] =
	{
		"invalid",		"bank 1",		"bank 2",		"bank 3",
		"bank 4",		"bank 5",		"bank 6",		"bank 7",
		"bank 8",		"bank 9",		"bank 10",		"bank 11",
		"bank 12",		"bank 13",		"bank 14",		"bank 15",
		"bank 16",		"bank 17",		"bank 18",		"bank 19",
		"bank 20",		"bank 21",		"bank 22",		"bank 23",
		"bank 24",		"RAM",			"ROM",			"RAMROM",
		"nop",			"unused 1",		"unused 2",		"unmapped"
	};

	int l1count = 1 << LEVEL1_BITS;
	int l2count = 1 << LEVEL2_BITS;
	int i, j;

	fprintf(file, "  Address bits = %d\n", space->abits);
	fprintf(file, "     Data bits = %d\n", space->dbits);
	fprintf(file, "       L1 bits = %d\n", LEVEL1_BITS);
	fprintf(file, "       L2 bits = %d\n", LEVEL2_BITS);
	fprintf(file, "  Address mask = %X\n", space->mask);
	fprintf(file, "\n");

	for (i = 0; i < l1count; i++)
	{
		UINT8 entry = table->table[i];
		if (entry != STATIC_UNMAP)
		{
			fprintf(file, "%05X  %08X-%08X    = %02X: ", i,
					i << LEVEL2_BITS,
					((i+1) << LEVEL2_BITS) - 1, entry);
			if (entry < STATIC_COUNT)
				fprintf(file, "%s [offset=%08X]\n", strings[entry], table->handlers[entry].offset);
			else if (entry < SUBTABLE_BASE)
				fprintf(file, "handler(%08X) [offset=%08X]\n", (UINT32)table->handlers[entry].handler, table->handlers[entry].offset);
			else
			{
				fprintf(file, "subtable %d\n", entry & SUBTABLE_MASK);
				entry &= SUBTABLE_MASK;

				for (j = 0; j < l2count; j++)
				{
					UINT8 entry2 = table->table[(1 << LEVEL1_BITS) + (entry << LEVEL2_BITS) + j];
					if (entry2 != STATIC_UNMAP)
					{
						fprintf(file, "   %05X  %08X-%08X = %02X: ", j,
								((i << LEVEL2_BITS) | j),
								((i << LEVEL2_BITS) | (j+1)) - 1, entry2);
						if (entry2 < STATIC_COUNT)
							fprintf(file, "%s [offset=%08X]\n", strings[entry2], table->handlers[entry2].offset);
						else if (entry2 < SUBTABLE_BASE)
							fprintf(file, "handler(%08X) [offset=%08X]\n", (UINT32)table->handlers[entry2].handler, table->handlers[entry2].offset);
						else
							fprintf(file, "subtable %d???????????\n", entry2 & SUBTABLE_MASK);
					}
				}
			}
		}
	}
}

static void mem_dump(void)
{
	FILE *file = fopen("memdump.log", "w");
	int cpunum, spacenum;

	/* skip if we can't open the file */
	if (!file)
		return;

	/* loop over CPUs */
	for (cpunum = 0; cpunum < cpu_gettotalcpu(); cpunum++)
		for (spacenum = 0; spacenum < ADDRESS_SPACES; spacenum++)
			if (cpudata[cpunum].space[spacenum].abits)
			{
				fprintf(file, "\n\n"
				              "=========================================\n"
				              "CPU %d address space %d read handler dump\n"
				              "=========================================\n", cpunum, spacenum);
				dump_map(file, &cpudata[cpunum].space[spacenum], &cpudata[cpunum].space[spacenum].read);

				fprintf(file, "\n\n"
				              "==========================================\n"
				              "CPU %d address space %d write handler dump\n"
				              "==========================================\n", cpunum, spacenum);
				dump_map(file, &cpudata[cpunum].space[spacenum], &cpudata[cpunum].space[spacenum].write);
			}
	fclose(file);
}

#endif
