/***************************************************************************

	memory.h

	Functions which handle the CPU memory accesses.

***************************************************************************/

#ifndef _MEMORY_H
#define _MEMORY_H

#include "osd_cpu.h"
#include <stddef.h>

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Versions of GNU C earlier that 2.7 appear to have problems with the
 * __attribute__ definition of UNUSEDARG, so we act as if it was not a
 * GNU compiler.
 */

#ifdef __GNUC__
#if (__GNUC__ < 2) || ((__GNUC__ == 2) && (__GNUC_MINOR__ <= 7))
#define UNUSEDARG
#else
#define UNUSEDARG __attribute__((__unused__))
#endif
#else
#define UNUSEDARG
#endif


/*
 * Use __builtin_expect on GNU C 3.0 and above
 */
#ifdef __GNUC__
#if (__GNUC__ < 3)
#define UNEXPECTED(exp)	(exp)
#else
#define UNEXPECTED(exp)	 __builtin_expect((exp), 0)
#endif
#else
#define UNEXPECTED(exp)	(exp)
#endif



/***************************************************************************

	Parameters

***************************************************************************/

#ifdef MAME_DEBUG
#define CPUREADOP_SAFETY_NONE		0
#define CPUREADOP_SAFETY_PARTIAL	0
#define CPUREADOP_SAFETY_FULL		1
#elif defined(MESS)
#define CPUREADOP_SAFETY_NONE		0
#define CPUREADOP_SAFETY_PARTIAL	1
#define CPUREADOP_SAFETY_FULL		0
#else
#define CPUREADOP_SAFETY_NONE		1
#define CPUREADOP_SAFETY_PARTIAL	0
#define CPUREADOP_SAFETY_FULL		0
#endif


	
/***************************************************************************

	Basic type definitions

	These types are used for memory handlers.

***************************************************************************/

/* ----- typedefs for data and offset types ----- */
typedef UINT8			data8_t;
typedef UINT16			data16_t;
typedef UINT32			data32_t;
typedef UINT64			data64_t;
typedef UINT32			offs_t;

/* ----- typedefs for the various common memory handlers ----- */
typedef data8_t			(*read8_handler)  (UNUSEDARG offs_t offset);
typedef void			(*write8_handler) (UNUSEDARG offs_t offset, UNUSEDARG data8_t data);
typedef data16_t		(*read16_handler) (UNUSEDARG offs_t offset, UNUSEDARG data16_t mem_mask);
typedef void			(*write16_handler)(UNUSEDARG offs_t offset, UNUSEDARG data16_t data, UNUSEDARG data16_t mem_mask);
typedef data32_t		(*read32_handler) (UNUSEDARG offs_t offset, UNUSEDARG data32_t mem_mask);
typedef void			(*write32_handler)(UNUSEDARG offs_t offset, UNUSEDARG data32_t data, UNUSEDARG data32_t mem_mask);
typedef data64_t		(*read64_handler) (UNUSEDARG offs_t offset, UNUSEDARG data64_t mem_mask);
typedef void			(*write64_handler)(UNUSEDARG offs_t offset, UNUSEDARG data64_t data, UNUSEDARG data64_t mem_mask);
typedef offs_t			(*opbase_handler) (UNUSEDARG offs_t address);

/* ----- typedefs for memory allocated by the memory system ----- */
struct memory_block_t
{
	UINT8			cpunum;
	UINT8			spacenum;
	offs_t 			start, end;
    UINT8 *			data;
};

struct memory_accessors_t
{
	data8_t			(*read_byte)(offs_t offset);
	data16_t		(*read_word)(offs_t offset);
	data32_t		(*read_dword)(offs_t offset);
	data64_t		(*read_qword)(offs_t offset);

	void			(*write_byte)(offs_t offset, data8_t data);
	void			(*write_word)(offs_t offset, data16_t data);
	void			(*write_dword)(offs_t offset, data32_t data);
	void			(*write_qword)(offs_t offset, data64_t data);
};



/***************************************************************************

	Basic macros

***************************************************************************/

/* ----- macros for declaring the various common memory handlers ----- */
#define READ8_HANDLER(name) 	data8_t  name(UNUSEDARG offs_t offset)
#define WRITE8_HANDLER(name) 	void     name(UNUSEDARG offs_t offset, UNUSEDARG data8_t data)
#define READ16_HANDLER(name)	data16_t name(UNUSEDARG offs_t offset, UNUSEDARG data16_t mem_mask)
#define WRITE16_HANDLER(name)	void     name(UNUSEDARG offs_t offset, UNUSEDARG data16_t data, UNUSEDARG data16_t mem_mask)
#define READ32_HANDLER(name)	data32_t name(UNUSEDARG offs_t offset, UNUSEDARG data32_t mem_mask)
#define WRITE32_HANDLER(name)	void     name(UNUSEDARG offs_t offset, UNUSEDARG data32_t data, UNUSEDARG data32_t mem_mask)
#define READ64_HANDLER(name)	data64_t name(UNUSEDARG offs_t offset, UNUSEDARG data64_t mem_mask)
#define WRITE64_HANDLER(name)	void     name(UNUSEDARG offs_t offset, UNUSEDARG data64_t data, UNUSEDARG data64_t mem_mask)
#define OPBASE_HANDLER(name)	offs_t   name(UNUSEDARG offs_t address)

/* ----- macros for accessing bytes and words within larger chunks ----- */
#ifdef LSB_FIRST
	#define BYTE_XOR_BE(a)  	((a) ^ 1)				/* read/write a byte to a 16-bit space */
	#define BYTE_XOR_LE(a)  	(a)
	#define BYTE4_XOR_BE(a) 	((a) ^ 3)				/* read/write a byte to a 32-bit space */
	#define BYTE4_XOR_LE(a) 	(a)
	#define WORD_XOR_BE(a)  	((a) ^ 2)				/* read/write a word to a 32-bit space */
	#define WORD_XOR_LE(a)  	(a)
	#define BYTE8_XOR_BE(a) 	((a) ^ 7)				/* read/write a byte to a 64-bit space */
	#define BYTE8_XOR_LE(a) 	(a)
	#define WORD2_XOR_BE(a)  	((a) ^ 6)				/* read/write a word to a 64-bit space */
	#define WORD2_XOR_LE(a)  	(a)
	#define DWORD_XOR_BE(a)  	((a) ^ 4)				/* read/write a dword to a 64-bit space */
	#define DWORD_XOR_LE(a)  	(a)
#else
	#define BYTE_XOR_BE(a)  	(a)
	#define BYTE_XOR_LE(a)  	((a) ^ 1)				/* read/write a byte to a 16-bit space */
	#define BYTE4_XOR_BE(a) 	(a)
	#define BYTE4_XOR_LE(a) 	((a) ^ 3)				/* read/write a byte to a 32-bit space */
	#define WORD_XOR_BE(a)  	(a)
	#define WORD_XOR_LE(a)  	((a) ^ 2)				/* read/write a word to a 32-bit space */
	#define BYTE8_XOR_BE(a) 	(a)
	#define BYTE8_XOR_LE(a) 	((a) ^ 7)				/* read/write a byte to a 64-bit space */
	#define WORD2_XOR_BE(a)  	(a)
	#define WORD2_XOR_LE(a)  	((a) ^ 6)				/* read/write a word to a 64-bit space */
	#define DWORD_XOR_BE(a)  	(a)
	#define DWORD_XOR_LE(a)  	((a) ^ 4)				/* read/write a dword to a 64-bit space */
#endif



/***************************************************************************

	Memory array constants

	These apply to values in the array of read/write handlers that is
	declared within each driver.

***************************************************************************/

/* ----- memory address space constants ----- */
#define MEMORY_SPACE_SHIFT		28						/* shift to get at the address space bits */
#define MEMORY_SPACE_MASK		(0x0f << MEMORY_SPACE_SHIFT) /* mask to get at the address space bits */
#define MEMORY_SPACE_PROGRAM	((ADDRESS_SPACE_PROGRAM+1) << MEMORY_SPACE_SHIFT)	/* this memory array is for program memory (and data for non-Harvard architectures) */
#define MEMORY_SPACE_DATA		((ADDRESS_SPACE_DATA+1) << MEMORY_SPACE_SHIFT) 		/* this memory array is for data memory (for Harvard architectures) */
#define MEMORY_SPACE_IO			((ADDRESS_SPACE_IO+1) << MEMORY_SPACE_SHIFT) 		/* this memory array is for I/O ports */

/* ----- memory direction constants ----- */
#define MEMORY_DIRECTION_MASK	0x03000000				/* mask to get at the direction bits */
#define MEMORY_DIRECTION_READ	0x01000000				/* this memory array is for reads */
#define MEMORY_DIRECTION_WRITE	0x02000000				/* this memory array is for writes */

/* ----- memory address bits constants ----- */
#define MEMORY_ABITS_MASK		0x00ff0000				/* set this bit to indicate the entry has address bits */
#define MEMORY_ABITS_SHIFT		16						/* shift to get the address bits count */
#define MEMORY_ABITS(n)			((n) << MEMORY_ABITS_SHIFT) /* macro to specify the number of address bits */

/* ----- memory unmapped value constants ----- */
#define MEMORY_UNMAPPED_MASK	0x00003000				/* mask to get at the unmapped value bits */
#define MEMORY_UNMAPPED_0		0x00001000				/* unmapped accesses to this memory array return all 0s */
#define MEMORY_UNMAPPED_1		0x00002000				/* unmapped accesses to this memory array return all 1s */

/* ----- memory width constants ----- */
#define MEMORY_WIDTH_MASK		0x00000007				/* mask to get at the width bits */
#define MEMORY_WIDTH_8			0x00000001				/* this memory array is for an 8-bit databus */
#define MEMORY_WIDTH_16 		0x00000002				/* this memory array is for a 16-bit databus */
#define MEMORY_WIDTH_32 		0x00000003				/* this memory array is for a 32-bit databus */
#define MEMORY_WIDTH_64 		0x00000004				/* this memory array is for a 64-bit databus */

/* ----- memory struct marker constants ----- */
#define MEMORY_MARKER			((offs_t)~0)			/* used in the end field to indicate end of array */

/* ----- static memory handler constants ----- */
#define STATIC_INVALID			0						/* invalid - should never be used */
#define STATIC_BANK1			1						/* banked memory #1 */
#define STATIC_BANK2			2						/* banked memory #2 */
#define STATIC_BANK3			3						/* banked memory #3 */
#define STATIC_BANK4			4						/* banked memory #4 */
#define STATIC_BANK5			5						/* banked memory #5 */
#define STATIC_BANK6			6						/* banked memory #6 */
#define STATIC_BANK7			7						/* banked memory #7 */
#define STATIC_BANK8			8						/* banked memory #8 */
#define STATIC_BANK9			9						/* banked memory #9 */
#define STATIC_BANK10			10						/* banked memory #10 */
#define STATIC_BANK11			11						/* banked memory #11 */
#define STATIC_BANK12			12						/* banked memory #12 */
#define STATIC_BANK13			13						/* banked memory #13 */
#define STATIC_BANK14			14						/* banked memory #14 */
#define STATIC_BANK15			15						/* banked memory #15 */
#define STATIC_BANK16			16						/* banked memory #16 */
#define STATIC_BANK17			17						/* banked memory #17 */
#define STATIC_BANK18			18						/* banked memory #18 */
#define STATIC_BANK19			19						/* banked memory #19 */
#define STATIC_BANK20			20						/* banked memory #20 */
#define STATIC_BANK21			21						/* banked memory #21 */
#define STATIC_BANK22			22						/* banked memory #22 */
#define STATIC_BANK23			23						/* banked memory #23 */
#define STATIC_BANK24			24						/* banked memory #24 */
#define STATIC_RAM				25						/* RAM - standard reads/writes */
#define STATIC_ROM				26						/* ROM - just like RAM, but writes to the bit-bucket */
#define STATIC_RAMROM			27						/* RAMROM - use for access in encrypted 8-bit systems */
#define STATIC_NOP				28						/* NOP - reads are 0, writes to the bit-bucket */
#define STATIC_UNUSED1			29						/* unused - reserved for future use */
#define STATIC_UNUSED2			30						/* unused - reserved for future use */
#define STATIC_UNMAP			31						/* unmapped - all unmapped memory goes here */
#define STATIC_COUNT			32						/* total number of static handlers */

/* ----- banking constants ----- */
#define MAX_BANKS				24						/* maximum number of banks */
#define STATIC_BANKMAX			(STATIC_RAM - 1)		/* handler constant of last bank */



/***************************************************************************

	Constants for static entries in memory read/write arrays

	The first 32 entries in the memory lookup table are reserved for
	"static" handlers. These are internal handlers for RAM, ROM, banks,
	and unmapped memory areas. The following definitions are the
	properly-casted versions of the STATIC_ constants above.

***************************************************************************/

/* 8-bit reads */
#define MRA8_BANK1				((read8_handler)STATIC_BANK1)
#define MRA8_BANK2				((read8_handler)STATIC_BANK2)
#define MRA8_BANK3				((read8_handler)STATIC_BANK3)
#define MRA8_BANK4				((read8_handler)STATIC_BANK4)
#define MRA8_BANK5				((read8_handler)STATIC_BANK5)
#define MRA8_BANK6				((read8_handler)STATIC_BANK6)
#define MRA8_BANK7				((read8_handler)STATIC_BANK7)
#define MRA8_BANK8				((read8_handler)STATIC_BANK8)
#define MRA8_BANK9				((read8_handler)STATIC_BANK9)
#define MRA8_BANK10				((read8_handler)STATIC_BANK10)
#define MRA8_BANK11				((read8_handler)STATIC_BANK11)
#define MRA8_BANK12				((read8_handler)STATIC_BANK12)
#define MRA8_BANK13				((read8_handler)STATIC_BANK13)
#define MRA8_BANK14				((read8_handler)STATIC_BANK14)
#define MRA8_BANK15				((read8_handler)STATIC_BANK15)
#define MRA8_BANK16				((read8_handler)STATIC_BANK16)
#define MRA8_BANK17				((read8_handler)STATIC_BANK17)
#define MRA8_BANK18				((read8_handler)STATIC_BANK18)
#define MRA8_BANK19				((read8_handler)STATIC_BANK19)
#define MRA8_BANK20				((read8_handler)STATIC_BANK20)
#define MRA8_BANK21				((read8_handler)STATIC_BANK21)
#define MRA8_BANK22				((read8_handler)STATIC_BANK22)
#define MRA8_BANK23				((read8_handler)STATIC_BANK23)
#define MRA8_BANK24				((read8_handler)STATIC_BANK24)
#define MRA8_NOP				((read8_handler)STATIC_NOP)
#define MRA8_RAM				((read8_handler)STATIC_RAM)
#define MRA8_ROM				((read8_handler)STATIC_ROM)
#define MRA8_RAMROM				((read8_handler)STATIC_RAMROM)

/* 8-bit writes */
#define MWA8_BANK1				((write8_handler)STATIC_BANK1)
#define MWA8_BANK2				((write8_handler)STATIC_BANK2)
#define MWA8_BANK3				((write8_handler)STATIC_BANK3)
#define MWA8_BANK4				((write8_handler)STATIC_BANK4)
#define MWA8_BANK5				((write8_handler)STATIC_BANK5)
#define MWA8_BANK6				((write8_handler)STATIC_BANK6)
#define MWA8_BANK7				((write8_handler)STATIC_BANK7)
#define MWA8_BANK8				((write8_handler)STATIC_BANK8)
#define MWA8_BANK9				((write8_handler)STATIC_BANK9)
#define MWA8_BANK10				((write8_handler)STATIC_BANK10)
#define MWA8_BANK11				((write8_handler)STATIC_BANK11)
#define MWA8_BANK12				((write8_handler)STATIC_BANK12)
#define MWA8_BANK13				((write8_handler)STATIC_BANK13)
#define MWA8_BANK14				((write8_handler)STATIC_BANK14)
#define MWA8_BANK15				((write8_handler)STATIC_BANK15)
#define MWA8_BANK16				((write8_handler)STATIC_BANK16)
#define MWA8_BANK17				((write8_handler)STATIC_BANK17)
#define MWA8_BANK18				((write8_handler)STATIC_BANK18)
#define MWA8_BANK19				((write8_handler)STATIC_BANK19)
#define MWA8_BANK20				((write8_handler)STATIC_BANK20)
#define MWA8_BANK21				((write8_handler)STATIC_BANK21)
#define MWA8_BANK22				((write8_handler)STATIC_BANK22)
#define MWA8_BANK23				((write8_handler)STATIC_BANK23)
#define MWA8_BANK24				((write8_handler)STATIC_BANK24)
#define MWA8_NOP				((write8_handler)STATIC_NOP)
#define MWA8_RAM				((write8_handler)STATIC_RAM)
#define MWA8_ROM				((write8_handler)STATIC_ROM)
#define MWA8_RAMROM				((write8_handler)STATIC_RAMROM)

/* 16-bit reads */
#define MRA16_BANK1				((read16_handler)STATIC_BANK1)
#define MRA16_BANK2				((read16_handler)STATIC_BANK2)
#define MRA16_BANK3				((read16_handler)STATIC_BANK3)
#define MRA16_BANK4				((read16_handler)STATIC_BANK4)
#define MRA16_BANK5				((read16_handler)STATIC_BANK5)
#define MRA16_BANK6				((read16_handler)STATIC_BANK6)
#define MRA16_BANK7				((read16_handler)STATIC_BANK7)
#define MRA16_BANK8				((read16_handler)STATIC_BANK8)
#define MRA16_BANK9				((read16_handler)STATIC_BANK9)
#define MRA16_BANK10			((read16_handler)STATIC_BANK10)
#define MRA16_BANK11			((read16_handler)STATIC_BANK11)
#define MRA16_BANK12			((read16_handler)STATIC_BANK12)
#define MRA16_BANK13			((read16_handler)STATIC_BANK13)
#define MRA16_BANK14			((read16_handler)STATIC_BANK14)
#define MRA16_BANK15			((read16_handler)STATIC_BANK15)
#define MRA16_BANK16			((read16_handler)STATIC_BANK16)
#define MRA16_BANK17			((read16_handler)STATIC_BANK17)
#define MRA16_BANK18			((read16_handler)STATIC_BANK18)
#define MRA16_BANK19			((read16_handler)STATIC_BANK19)
#define MRA16_BANK20			((read16_handler)STATIC_BANK20)
#define MRA16_BANK21			((read16_handler)STATIC_BANK21)
#define MRA16_BANK22			((read16_handler)STATIC_BANK22)
#define MRA16_BANK23			((read16_handler)STATIC_BANK23)
#define MRA16_BANK24			((read16_handler)STATIC_BANK24)
#define MRA16_NOP				((read16_handler)STATIC_NOP)
#define MRA16_RAM				((read16_handler)STATIC_RAM)
#define MRA16_ROM				((read16_handler)STATIC_ROM)

/* 16-bit writes */
#define MWA16_BANK1				((write16_handler)STATIC_BANK1)
#define MWA16_BANK2				((write16_handler)STATIC_BANK2)
#define MWA16_BANK3				((write16_handler)STATIC_BANK3)
#define MWA16_BANK4				((write16_handler)STATIC_BANK4)
#define MWA16_BANK5				((write16_handler)STATIC_BANK5)
#define MWA16_BANK6				((write16_handler)STATIC_BANK6)
#define MWA16_BANK7				((write16_handler)STATIC_BANK7)
#define MWA16_BANK8				((write16_handler)STATIC_BANK8)
#define MWA16_BANK9				((write16_handler)STATIC_BANK9)
#define MWA16_BANK10			((write16_handler)STATIC_BANK10)
#define MWA16_BANK11			((write16_handler)STATIC_BANK11)
#define MWA16_BANK12			((write16_handler)STATIC_BANK12)
#define MWA16_BANK13			((write16_handler)STATIC_BANK13)
#define MWA16_BANK14			((write16_handler)STATIC_BANK14)
#define MWA16_BANK15			((write16_handler)STATIC_BANK15)
#define MWA16_BANK16			((write16_handler)STATIC_BANK16)
#define MWA16_BANK17			((write16_handler)STATIC_BANK17)
#define MWA16_BANK18			((write16_handler)STATIC_BANK18)
#define MWA16_BANK19			((write16_handler)STATIC_BANK19)
#define MWA16_BANK20			((write16_handler)STATIC_BANK20)
#define MWA16_BANK21			((write16_handler)STATIC_BANK21)
#define MWA16_BANK22			((write16_handler)STATIC_BANK22)
#define MWA16_BANK23			((write16_handler)STATIC_BANK23)
#define MWA16_BANK24			((write16_handler)STATIC_BANK24)
#define MWA16_NOP				((write16_handler)STATIC_NOP)
#define MWA16_RAM				((write16_handler)STATIC_RAM)
#define MWA16_ROM				((write16_handler)STATIC_ROM)

/* 32-bit reads */
#define MRA32_BANK1				((read32_handler)STATIC_BANK1)
#define MRA32_BANK2				((read32_handler)STATIC_BANK2)
#define MRA32_BANK3				((read32_handler)STATIC_BANK3)
#define MRA32_BANK4				((read32_handler)STATIC_BANK4)
#define MRA32_BANK5				((read32_handler)STATIC_BANK5)
#define MRA32_BANK6				((read32_handler)STATIC_BANK6)
#define MRA32_BANK7				((read32_handler)STATIC_BANK7)
#define MRA32_BANK8				((read32_handler)STATIC_BANK8)
#define MRA32_BANK9				((read32_handler)STATIC_BANK9)
#define MRA32_BANK10			((read32_handler)STATIC_BANK10)
#define MRA32_BANK11			((read32_handler)STATIC_BANK11)
#define MRA32_BANK12			((read32_handler)STATIC_BANK12)
#define MRA32_BANK13			((read32_handler)STATIC_BANK13)
#define MRA32_BANK14			((read32_handler)STATIC_BANK14)
#define MRA32_BANK15			((read32_handler)STATIC_BANK15)
#define MRA32_BANK16			((read32_handler)STATIC_BANK16)
#define MRA32_BANK17			((read32_handler)STATIC_BANK17)
#define MRA32_BANK18			((read32_handler)STATIC_BANK18)
#define MRA32_BANK19			((read32_handler)STATIC_BANK19)
#define MRA32_BANK20			((read32_handler)STATIC_BANK20)
#define MRA32_BANK21			((read32_handler)STATIC_BANK21)
#define MRA32_BANK22			((read32_handler)STATIC_BANK22)
#define MRA32_BANK23			((read32_handler)STATIC_BANK23)
#define MRA32_BANK24			((read32_handler)STATIC_BANK24)
#define MRA32_NOP				((read32_handler)STATIC_NOP)
#define MRA32_RAM				((read32_handler)STATIC_RAM)
#define MRA32_ROM				((read32_handler)STATIC_ROM)

/* 32-bit writes */
#define MWA32_BANK1				((write32_handler)STATIC_BANK1)
#define MWA32_BANK2				((write32_handler)STATIC_BANK2)
#define MWA32_BANK3				((write32_handler)STATIC_BANK3)
#define MWA32_BANK4				((write32_handler)STATIC_BANK4)
#define MWA32_BANK5				((write32_handler)STATIC_BANK5)
#define MWA32_BANK6				((write32_handler)STATIC_BANK6)
#define MWA32_BANK7				((write32_handler)STATIC_BANK7)
#define MWA32_BANK8				((write32_handler)STATIC_BANK8)
#define MWA32_BANK9				((write32_handler)STATIC_BANK9)
#define MWA32_BANK10			((write32_handler)STATIC_BANK10)
#define MWA32_BANK11			((write32_handler)STATIC_BANK11)
#define MWA32_BANK12			((write32_handler)STATIC_BANK12)
#define MWA32_BANK13			((write32_handler)STATIC_BANK13)
#define MWA32_BANK14			((write32_handler)STATIC_BANK14)
#define MWA32_BANK15			((write32_handler)STATIC_BANK15)
#define MWA32_BANK16			((write32_handler)STATIC_BANK16)
#define MWA32_BANK17			((write32_handler)STATIC_BANK17)
#define MWA32_BANK18			((write32_handler)STATIC_BANK18)
#define MWA32_BANK19			((write32_handler)STATIC_BANK19)
#define MWA32_BANK20			((write32_handler)STATIC_BANK20)
#define MWA32_BANK21			((write32_handler)STATIC_BANK21)
#define MWA32_BANK22			((write32_handler)STATIC_BANK22)
#define MWA32_BANK23			((write32_handler)STATIC_BANK23)
#define MWA32_BANK24			((write32_handler)STATIC_BANK24)
#define MWA32_NOP				((write32_handler)STATIC_NOP)
#define MWA32_RAM				((write32_handler)STATIC_RAM)
#define MWA32_ROM				((write32_handler)STATIC_ROM)

/* 64-bit reads */
#define MRA64_BANK1				((read64_handler)STATIC_BANK1)
#define MRA64_BANK2				((read64_handler)STATIC_BANK2)
#define MRA64_BANK3				((read64_handler)STATIC_BANK3)
#define MRA64_BANK4				((read64_handler)STATIC_BANK4)
#define MRA64_BANK5				((read64_handler)STATIC_BANK5)
#define MRA64_BANK6				((read64_handler)STATIC_BANK6)
#define MRA64_BANK7				((read64_handler)STATIC_BANK7)
#define MRA64_BANK8				((read64_handler)STATIC_BANK8)
#define MRA64_BANK9				((read64_handler)STATIC_BANK9)
#define MRA64_BANK10			((read64_handler)STATIC_BANK10)
#define MRA64_BANK11			((read64_handler)STATIC_BANK11)
#define MRA64_BANK12			((read64_handler)STATIC_BANK12)
#define MRA64_BANK13			((read64_handler)STATIC_BANK13)
#define MRA64_BANK14			((read64_handler)STATIC_BANK14)
#define MRA64_BANK15			((read64_handler)STATIC_BANK15)
#define MRA64_BANK16			((read64_handler)STATIC_BANK16)
#define MRA64_BANK17			((read64_handler)STATIC_BANK17)
#define MRA64_BANK18			((read64_handler)STATIC_BANK18)
#define MRA64_BANK19			((read64_handler)STATIC_BANK19)
#define MRA64_BANK20			((read64_handler)STATIC_BANK20)
#define MRA64_BANK21			((read64_handler)STATIC_BANK21)
#define MRA64_BANK22			((read64_handler)STATIC_BANK22)
#define MRA64_BANK23			((read64_handler)STATIC_BANK23)
#define MRA64_BANK24			((read64_handler)STATIC_BANK24)
#define MRA64_NOP				((read64_handler)STATIC_NOP)
#define MRA64_RAM				((read64_handler)STATIC_RAM)
#define MRA64_ROM				((read64_handler)STATIC_ROM)

/* 64-bit writes */
#define MWA64_BANK1				((write64_handler)STATIC_BANK1)
#define MWA64_BANK2				((write64_handler)STATIC_BANK2)
#define MWA64_BANK3				((write64_handler)STATIC_BANK3)
#define MWA64_BANK4				((write64_handler)STATIC_BANK4)
#define MWA64_BANK5				((write64_handler)STATIC_BANK5)
#define MWA64_BANK6				((write64_handler)STATIC_BANK6)
#define MWA64_BANK7				((write64_handler)STATIC_BANK7)
#define MWA64_BANK8				((write64_handler)STATIC_BANK8)
#define MWA64_BANK9				((write64_handler)STATIC_BANK9)
#define MWA64_BANK10			((write64_handler)STATIC_BANK10)
#define MWA64_BANK11			((write64_handler)STATIC_BANK11)
#define MWA64_BANK12			((write64_handler)STATIC_BANK12)
#define MWA64_BANK13			((write64_handler)STATIC_BANK13)
#define MWA64_BANK14			((write64_handler)STATIC_BANK14)
#define MWA64_BANK15			((write64_handler)STATIC_BANK15)
#define MWA64_BANK16			((write64_handler)STATIC_BANK16)
#define MWA64_BANK17			((write64_handler)STATIC_BANK17)
#define MWA64_BANK18			((write64_handler)STATIC_BANK18)
#define MWA64_BANK19			((write64_handler)STATIC_BANK19)
#define MWA64_BANK20			((write64_handler)STATIC_BANK20)
#define MWA64_BANK21			((write64_handler)STATIC_BANK21)
#define MWA64_BANK22			((write64_handler)STATIC_BANK22)
#define MWA64_BANK23			((write64_handler)STATIC_BANK23)
#define MWA64_BANK24			((write64_handler)STATIC_BANK24)
#define MWA64_NOP				((write64_handler)STATIC_NOP)
#define MWA64_RAM				((write64_handler)STATIC_RAM)
#define MWA64_ROM				((write64_handler)STATIC_ROM)



/***************************************************************************

	Memory array type definitions

	Note that the memory hooks are not passed the actual memory address
	where the operation takes place, but the offset from the beginning
	of the block they are assigned to. This makes handling of mirror
	addresses easier, and makes the handlers a bit more "object oriented".
	If you handler needs to read/write the main memory area, provide a
	"base" pointer: it will be initialized by the main engine to point to
	the beginning of the memory block assigned to the handler. You may
	also provided a pointer to "size": it will be set to the length of
	the memory area processed by the handler.

***************************************************************************/

/* ----- structs for memory read arrays ----- */
struct memory_read8_map_t
{
	offs_t				start, end, mask;	/* start/end pair, or match/select/mask triplet */
	read8_handler 		handler;			/* handler callback */
};

struct memory_read16_map_t
{
	offs_t				start, end, mask;	/* start/end pair, or match/select/mask triplet */
	read16_handler 		handler;			/* handler callback */
};

struct memory_read32_map_t
{
	offs_t				start, end, mask;	/* start/end pair, or match/select/mask triplet */
	read32_handler		handler;			/* handler callback */
};

struct memory_read64_map_t
{
	offs_t				start, end, mask;	/* start/end pair, or match/select/mask triplet */
	read64_handler		handler;			/* handler callback */
};

/* ----- structs for memory write arrays ----- */
struct memory_write8_map_t
{
	offs_t				start, end, mask;	/* start/end pair, or match/select/mask triplet */
	write8_handler		handler;			/* handler callback */
	data8_t **			base;				/* receives pointer to memory (optional) */
    size_t *			size;				/* receives size of memory in bytes (optional) */
};

struct memory_write16_map_t
{
	offs_t				start, end, mask;	/* start/end pair, or match/select/mask triplet */
	write16_handler 	handler;			/* handler callback */
	data16_t **			base;				/* receives pointer to memory (optional) */
    size_t *			size;				/* receives size of memory in bytes (optional) */
};

struct memory_write32_map_t
{
	offs_t				start, end, mask;	/* start/end pair, or match/select/mask triplet */
	write32_handler 	handler;			/* handler callback */
	data32_t **			base;				/* receives pointer to memory (optional) */
	size_t *			size;				/* receives size of memory in bytes (optional) */
};

struct memory_write64_map_t
{
	offs_t				start, end, mask;	/* start/end pair, or match/select/mask triplet */
	write64_handler 	handler;			/* handler callback */
	data64_t **			base;				/* receives pointer to memory (optional) */
	size_t *			size;				/* receives size of memory in bytes (optional) */
};

/* ----- structs to contain internal data ----- */
struct address_space_t
{
	offs_t				addrmask;			/* address mask */
	UINT8 *				readlookup;			/* read table lookup */
	UINT8 *				writelookup;		/* write table lookup */
	struct handler_data *readhandlers;		/* read handlers */
	struct handler_data *writehandlers;		/* write handlers */
	struct memory_accessors_t *accessors;	/* pointers to the memory accessors */ 
};



/***************************************************************************

	Memory/port array macros

***************************************************************************/

/* ----- macros for identifying memory/port struct markers ----- */
#define IS_MEMORY_MARKER(ma)				((ma)->start == MEMORY_MARKER && (ma)->end < MEMORY_MARKER)
#define IS_MEMORY_END(ma)					((ma)->start == MEMORY_MARKER && (ma)->end == 0)

/* ----- macros for defining the start/stop points ----- */
#define MEMORY_MAP_START(t,n,f)				const struct t n[] = { { MEMORY_MARKER, (f) },
#define MEMORY_MAP_FLAGS(f)					{ MEMORY_MARKER, (f) },
#define MEMORY_MAP_END						{ MEMORY_MARKER, 0 } };

/* ----- macros for declaring the start of a memory struct array ----- */
#define MEMORY_READ8_MAP_START(name,f)		MEMORY_MAP_START(memory_read8_map_t,   name, MEMORY_DIRECTION_READ  | MEMORY_WIDTH_8  | (f))
#define MEMORY_WRITE8_MAP_START(name,f)		MEMORY_MAP_START(memory_write8_map_t,  name, MEMORY_DIRECTION_WRITE | MEMORY_WIDTH_8  | (f))
#define MEMORY_READ16_MAP_START(name,f)		MEMORY_MAP_START(memory_read16_map_t,  name, MEMORY_DIRECTION_READ  | MEMORY_WIDTH_16 | (f))
#define MEMORY_WRITE16_MAP_START(name,f)	MEMORY_MAP_START(memory_write16_map_t, name, MEMORY_DIRECTION_WRITE | MEMORY_WIDTH_16 | (f))
#define MEMORY_READ32_MAP_START(name,f)		MEMORY_MAP_START(memory_read32_map_t,  name, MEMORY_DIRECTION_READ  | MEMORY_WIDTH_32 | (f))
#define MEMORY_WRITE32_MAP_START(name,f)	MEMORY_MAP_START(memory_write32_map_t, name, MEMORY_DIRECTION_WRITE | MEMORY_WIDTH_32 | (f))
#define MEMORY_READ64_MAP_START(name,f)		MEMORY_MAP_START(memory_read64_map_t,  name, MEMORY_DIRECTION_READ  | MEMORY_WIDTH_64 | (f))
#define MEMORY_WRITE64_MAP_START(name,f)	MEMORY_MAP_START(memory_write64_map_t, name, MEMORY_DIRECTION_WRITE | MEMORY_WIDTH_64 | (f))

/* ----- macros for declaring the start of a program memory struct array ----- */
#define PROGRAM_READ8_MAP_START(name)		MEMORY_READ8_MAP_START(name, MEMORY_SPACE_PROGRAM)
#define PROGRAM_WRITE8_MAP_START(name)		MEMORY_WRITE8_MAP_START(name, MEMORY_SPACE_PROGRAM)
#define PROGRAM_READ16_MAP_START(name)		MEMORY_READ16_MAP_START(name, MEMORY_SPACE_PROGRAM)
#define PROGRAM_WRITE16_MAP_START(name)		MEMORY_WRITE16_MAP_START(name, MEMORY_SPACE_PROGRAM)
#define PROGRAM_READ32_MAP_START(name)		MEMORY_READ32_MAP_START(name, MEMORY_SPACE_PROGRAM)
#define PROGRAM_WRITE32_MAP_START(name)		MEMORY_WRITE32_MAP_START(name, MEMORY_SPACE_PROGRAM)
#define PROGRAM_READ64_MAP_START(name)		MEMORY_READ64_MAP_START(name, MEMORY_SPACE_PROGRAM)
#define PROGRAM_WRITE64_MAP_START(name)		MEMORY_WRITE64_MAP_START(name, MEMORY_SPACE_PROGRAM)
#define PROGRAM_MAP_FLAGS(f)				MEMORY_MAP_FLAGS(f)
#define PROGRAM_MAP_END						MEMORY_MAP_END

/* ----- macros for declaring the start of a data memory struct array ----- */
#define DATA_READ8_MAP_START(name)			MEMORY_READ8_MAP_START(name, MEMORY_SPACE_DATA)
#define DATA_WRITE8_MAP_START(name)			MEMORY_WRITE8_MAP_START(name, MEMORY_SPACE_DATA)
#define DATA_READ16_MAP_START(name)			MEMORY_READ16_MAP_START(name, MEMORY_SPACE_DATA)
#define DATA_WRITE16_MAP_START(name)		MEMORY_WRITE16_MAP_START(name, MEMORY_SPACE_DATA)
#define DATA_READ32_MAP_START(name)			MEMORY_READ32_MAP_START(name, MEMORY_SPACE_DATA)
#define DATA_WRITE32_MAP_START(name)		MEMORY_WRITE32_MAP_START(name, MEMORY_SPACE_DATA)
#define DATA_READ64_MAP_START(name)			MEMORY_READ64_MAP_START(name, MEMORY_SPACE_DATA)
#define DATA_WRITE64_MAP_START(name)		MEMORY_WRITE64_MAP_START(name, MEMORY_SPACE_DATA)
#define DATA_MAP_FLAGS(f)					MEMORY_MAP_FLAGS(f)
#define DATA_MAP_END						MEMORY_MAP_END

/* ----- macros for declaring the start of a port struct array ----- */
#define IO_READ8_MAP_START(name)			MEMORY_READ8_MAP_START(name, MEMORY_SPACE_IO)
#define IO_WRITE8_MAP_START(name)			MEMORY_WRITE8_MAP_START(name, MEMORY_SPACE_IO)
#define IO_READ16_MAP_START(name)			MEMORY_READ16_MAP_START(name, MEMORY_SPACE_IO)
#define IO_WRITE16_MAP_START(name)			MEMORY_WRITE16_MAP_START(name, MEMORY_SPACE_IO)
#define IO_READ32_MAP_START(name)			MEMORY_READ32_MAP_START(name, MEMORY_SPACE_IO)
#define IO_WRITE32_MAP_START(name)			MEMORY_WRITE32_MAP_START(name, MEMORY_SPACE_IO)
#define IO_READ64_MAP_START(name)			MEMORY_READ64_MAP_START(name, MEMORY_SPACE_IO)
#define IO_WRITE64_MAP_START(name)			MEMORY_WRITE64_MAP_START(name, MEMORY_SPACE_IO)
#define IO_MAP_FLAGS(f)						MEMORY_MAP_FLAGS(f)
#define IO_MAP_END							MEMORY_MAP_END

/* ----- macros for declaring address ranges (start, end pairs) ----- */
#define READ_RANGE(a,b,c)					{ (a), (b), 0, (c) },
#define WRITE_RANGE(a,b,c)					{ (a), (b), 0, (c) },
#define WRITE_RANGE_P(a,b,c,d)				{ (a), (b), 0, (c), (d) },
#define WRITE_RANGE_PS(a,b,c,d,e)			{ (a), (b), 0, (c), (d), (e) },

/* ----- macros for declaring address spaces (match, select, mask triplets) ----- */
#define READ_SPACE(a,b,c,d)					{ (a), (b), (c), (d) },
#define WRITE_SPACE(a,b,c,d)				{ (a), (b), (c), (d) },
#define WRITE_SPACE_P(a,b,c,d,e)			{ (a), (b), (c), (d), (e) },
#define WRITE_SPACE_PS(a,b,c,d,e,f) 		{ (a), (b), (c), (d), (e), (f) },



/***************************************************************************

	Memory/port lookup constants

	These apply to values in the internal lookup table.

***************************************************************************/

/* ----- address spaces ----- */
#define ADDRESS_SPACES			3						/* maximum number of address spaces */
#define ADDRESS_SPACE_PROGRAM	0						/* program memory */
#define ADDRESS_SPACE_DATA		1						/* data memory */
#define ADDRESS_SPACE_IO		2						/* I/O ports */

/* ----- memory/port lookup table definitions ----- */
#define SUBTABLE_COUNT			64						/* number of slots reserved for subtables */
#define SUBTABLE_MASK			(SUBTABLE_COUNT-1)		/* mask to get at the subtable index */
#define SUBTABLE_BASE			(256-SUBTABLE_COUNT)	/* first index of a subtable */
#define ENTRY_COUNT				(SUBTABLE_BASE)			/* number of legitimate (non-subtable) entries */
#define SUBTABLE_ALLOC			8						/* number of subtables to allocate at a time */

/* ----- bit counts ----- */
#define SPARSE_THRESH			20						/* number of address bits above which we use sparse memory */
#define LEVEL1_BITS				18						/* number of address bits in the level 1 table */
#define LEVEL2_BITS				(32 - LEVEL1_BITS)		/* number of address bits in the level 2 table */

/* ----- external memory constants ----- */
#define MAX_EXT_MEMORY			64						/* maximum external memory areas we can allocate */



/***************************************************************************

	Memory/port lookup macros

	These are used for accessing the internal lookup table.

***************************************************************************/

/* ----- table lookup helpers ----- */
#define LEVEL1_INDEX(a)			((a) >> LEVEL2_BITS)
#define LEVEL2_INDEX(e,a)		((1 << LEVEL1_BITS) + (((e) & SUBTABLE_MASK) << LEVEL2_BITS) + ((a) & ((1 << LEVEL2_BITS) - 1)))

/* ----- sparse memory space detection ----- */
#define IS_SPARSE(a)			((a) > SPARSE_THRESH)



/***************************************************************************

	Function prototypes for core readmem/writemem routines

***************************************************************************/

/* ----- declare program memory handlers ----- */
data8_t program_read_byte_8(offs_t address);
void program_write_byte_8(offs_t address, data8_t data);

data8_t program_read_byte_16be(offs_t address);
data16_t program_read_word_16be(offs_t address);
void program_write_byte_16be(offs_t address, data8_t data);
void program_write_word_16be(offs_t address, data16_t data);
   
data8_t program_read_byte_16le(offs_t address);
data16_t program_read_word_16le(offs_t address);
void program_write_byte_16le(offs_t address, data8_t data);
void program_write_word_16le(offs_t address, data16_t data);

data8_t program_read_byte_32be(offs_t address);
data16_t program_read_word_32be(offs_t address);
data32_t program_read_dword_32be(offs_t address);
void program_write_byte_32be(offs_t address, data8_t data);
void program_write_word_32be(offs_t address, data16_t data);
void program_write_dword_32be(offs_t address, data32_t data);
   
data8_t program_read_byte_32le(offs_t address);
data16_t program_read_word_32le(offs_t address);
data32_t program_read_dword_32le(offs_t address);
void program_write_byte_32le(offs_t address, data8_t data);
void program_write_word_32le(offs_t address, data16_t data);
void program_write_dword_32le(offs_t address, data32_t data);
   
data8_t program_read_byte_64be(offs_t address);
data16_t program_read_word_64be(offs_t address);
data32_t program_read_dword_64be(offs_t address);
data64_t program_read_qword_64be(offs_t address);
void program_write_byte_64be(offs_t address, data8_t data);
void program_write_word_64be(offs_t address, data16_t data);
void program_write_dword_64be(offs_t address, data32_t data);
void program_write_qword_64be(offs_t address, data64_t data);
   
data8_t program_read_byte_64le(offs_t address);
data16_t program_read_word_64le(offs_t address);
data32_t program_read_dword_64le(offs_t address);
data64_t program_read_qword_64le(offs_t address);
void program_write_byte_64le(offs_t address, data8_t data);
void program_write_word_64le(offs_t address, data16_t data);
void program_write_dword_64le(offs_t address, data32_t data);
void program_write_qword_64le(offs_t address, data64_t data);

/* ----- declare data memory handlers ----- */
data8_t data_read_byte_8(offs_t address);
void data_write_byte_8(offs_t address, data8_t data);

data8_t data_read_byte_16be(offs_t address);
data16_t data_read_word_16be(offs_t address);
void data_write_byte_16be(offs_t address, data8_t data);
void data_write_word_16be(offs_t address, data16_t data);
   
data8_t data_read_byte_16le(offs_t address);
data16_t data_read_word_16le(offs_t address);
void data_write_byte_16le(offs_t address, data8_t data);
void data_write_word_16le(offs_t address, data16_t data);

data8_t data_read_byte_32be(offs_t address);
data16_t data_read_word_32be(offs_t address);
data32_t data_read_dword_32be(offs_t address);
void data_write_byte_32be(offs_t address, data8_t data);
void data_write_word_32be(offs_t address, data16_t data);
void data_write_dword_32be(offs_t address, data32_t data);
   
data8_t data_read_byte_32le(offs_t address);
data16_t data_read_word_32le(offs_t address);
data32_t data_read_dword_32le(offs_t address);
void data_write_byte_32le(offs_t address, data8_t data);
void data_write_word_32le(offs_t address, data16_t data);
void data_write_dword_32le(offs_t address, data32_t data);
   
data8_t data_read_byte_64be(offs_t address);
data16_t data_read_word_64be(offs_t address);
data32_t data_read_dword_64be(offs_t address);
data64_t data_read_qword_64be(offs_t address);
void data_write_byte_64be(offs_t address, data8_t data);
void data_write_word_64be(offs_t address, data16_t data);
void data_write_dword_64be(offs_t address, data32_t data);
void data_write_qword_64be(offs_t address, data64_t data);
   
data8_t data_read_byte_64le(offs_t address);
data16_t data_read_word_64le(offs_t address);
data32_t data_read_dword_64le(offs_t address);
data64_t data_read_qword_64le(offs_t address);
void data_write_byte_64le(offs_t address, data8_t data);
void data_write_word_64le(offs_t address, data16_t data);
void data_write_dword_64le(offs_t address, data32_t data);
void data_write_qword_64le(offs_t address, data64_t data);

/* ----- declare I/O memory handlers ----- */
data8_t io_read_byte_8(offs_t address);
void io_write_byte_8(offs_t address, data8_t data);

data8_t io_read_byte_16be(offs_t address);
data16_t io_read_word_16be(offs_t address);
void io_write_byte_16be(offs_t address, data8_t data);
void io_write_word_16be(offs_t address, data16_t data);
   
data8_t io_read_byte_16le(offs_t address);
data16_t io_read_word_16le(offs_t address);
void io_write_byte_16le(offs_t address, data8_t data);
void io_write_word_16le(offs_t address, data16_t data);

data8_t io_read_byte_32be(offs_t address);
data16_t io_read_word_32be(offs_t address);
data32_t io_read_dword_32be(offs_t address);
void io_write_byte_32be(offs_t address, data8_t data);
void io_write_word_32be(offs_t address, data16_t data);
void io_write_dword_32be(offs_t address, data32_t data);
   
data8_t io_read_byte_32le(offs_t address);
data16_t io_read_word_32le(offs_t address);
data32_t io_read_dword_32le(offs_t address);
void io_write_byte_32le(offs_t address, data8_t data);
void io_write_word_32le(offs_t address, data16_t data);
void io_write_dword_32le(offs_t address, data32_t data);
   
data8_t io_read_byte_64be(offs_t address);
data16_t io_read_word_64be(offs_t address);
data32_t io_read_dword_64be(offs_t address);
data64_t io_read_qword_64be(offs_t address);
void io_write_byte_64be(offs_t address, data8_t data);
void io_write_word_64be(offs_t address, data16_t data);
void io_write_dword_64be(offs_t address, data32_t data);
void io_write_qword_64be(offs_t address, data64_t data);
   
data8_t io_read_byte_64le(offs_t address);
data16_t io_read_word_64le(offs_t address);
data32_t io_read_dword_64le(offs_t address);
data64_t io_read_qword_64le(offs_t address);
void io_write_byte_64le(offs_t address, data8_t data);
void io_write_word_64le(offs_t address, data16_t data);
void io_write_dword_64le(offs_t address, data32_t data);
void io_write_qword_64le(offs_t address, data64_t data);



/***************************************************************************

	Function prototypes for core memory functions

***************************************************************************/

/* ----- memory setup function ----- */
int			memory_init(void);
void		memory_shutdown(void);
void		memory_set_context(int activecpu);

/* ----- opcode base control ---- */
opbase_handler memory_set_opbase_handler(int cpunum, opbase_handler function);
void		memory_set_opbase(offs_t offset);

/* ----- separate opcode/data encryption helpers ---- */
void		memory_set_opcode_base(int cpunum, void *base);

/* ----- return a base pointer to memory ---- */
void *		memory_get_read_ptr(int cpunum, int spacenum, offs_t offset);
void *		memory_get_write_ptr(int cpunum, int spacenum, offs_t offset);

/* ----- dynamic memory mapping ----- */
data8_t *	memory_install_read8_handler  (int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, read8_handler handler);
data16_t *	memory_install_read16_handler (int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, read16_handler handler);
data32_t *	memory_install_read32_handler (int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, read32_handler handler);
data64_t *	memory_install_read64_handler (int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, read64_handler handler);
data8_t *	memory_install_write8_handler (int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, write8_handler handler);
data16_t *	memory_install_write16_handler(int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, write16_handler handler);
data32_t *	memory_install_write32_handler(int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, write32_handler handler);
data64_t *	memory_install_write64_handler(int cpunum, int spacenum, offs_t start, offs_t end, offs_t mask, write64_handler handler);



/***************************************************************************

	Global variables

***************************************************************************/

extern UINT8 			opcode_entry;		/* current entry for opcode fetching */
extern UINT8 *			OP_ROM;				/* opcode ROM base */
extern UINT8 *			OP_RAM;				/* opcode RAM base */
extern offs_t			OP_MEM_MIN;			/* opcode memory minimum */
extern offs_t			OP_MEM_MAX;			/* opcode memory maximum */
extern UINT8 *			cpu_bankbase[];		/* array of bank bases */
extern struct address_space_t address_space[];/* memory address spaces */
extern struct memory_block_t ext_memory[];		/* externally-allocated memory */



/***************************************************************************

	Helper macros and inlines

***************************************************************************/

/* ----- 16/32-bit memory accessing ----- */
#define COMBINE_DATA(varptr)		(*(varptr) = (*(varptr) & mem_mask) | (data & ~mem_mask))

/* ----- 16-bit memory accessing ----- */
#define ACCESSING_LSB16				((mem_mask & 0x00ff) == 0)
#define ACCESSING_MSB16				((mem_mask & 0xff00) == 0)
#define ACCESSING_LSB				ACCESSING_LSB16
#define ACCESSING_MSB				ACCESSING_MSB16

/* ----- 32-bit memory accessing ----- */
#define ACCESSING_LSW32				((mem_mask & 0x0000ffff) == 0)
#define ACCESSING_MSW32				((mem_mask & 0xffff0000) == 0)
#define ACCESSING_LSB32				((mem_mask & 0x000000ff) == 0)
#define ACCESSING_MSB32				((mem_mask & 0xff000000) == 0)

/* ----- opcode range safety checks ----- */
#if CPUREADOP_SAFETY_NONE
#define address_is_unsafe(A)		(0)
#elif CPUREADOP_SAFETY_PARTIAL
#define address_is_unsafe(A)		(UNEXPECTED((A) > OP_MEM_MAX))
#elif CPUREADOP_SAFETY_FULL
#define address_is_unsafe(A)		((UNEXPECTED((A) < OP_MEM_MIN) || UNEXPECTED((A) > OP_MEM_MAX)))
#else
#error Must set either CPUREADOP_SAFETY_NONE, CPUREADOP_SAFETY_PARTIAL or CPUREADOP_SAFETY_FULL
#endif

/* ----- generic memory access ----- */
INLINE data8_t  program_read_byte (offs_t offset) { return (*address_space[ADDRESS_SPACE_PROGRAM].accessors->read_byte)(offset); }
INLINE data16_t program_read_word (offs_t offset) { return (*address_space[ADDRESS_SPACE_PROGRAM].accessors->read_word)(offset); }
INLINE data32_t program_read_dword(offs_t offset) { return (*address_space[ADDRESS_SPACE_PROGRAM].accessors->read_dword)(offset); }
INLINE data64_t program_read_qword(offs_t offset) { return (*address_space[ADDRESS_SPACE_PROGRAM].accessors->read_qword)(offset); }

INLINE void	program_write_byte (offs_t offset, data8_t  data) { (*address_space[ADDRESS_SPACE_PROGRAM].accessors->write_byte)(offset, data); }
INLINE void	program_write_word (offs_t offset, data16_t data) { (*address_space[ADDRESS_SPACE_PROGRAM].accessors->write_word)(offset, data); }
INLINE void	program_write_dword(offs_t offset, data32_t data) { (*address_space[ADDRESS_SPACE_PROGRAM].accessors->write_dword)(offset, data); }
INLINE void	program_write_qword(offs_t offset, data64_t data) { (*address_space[ADDRESS_SPACE_PROGRAM].accessors->write_qword)(offset, data); }

INLINE data8_t  data_read_byte (offs_t offset) { return (*address_space[ADDRESS_SPACE_DATA].accessors->read_byte)(offset); }
INLINE data16_t data_read_word (offs_t offset) { return (*address_space[ADDRESS_SPACE_DATA].accessors->read_word)(offset); }
INLINE data32_t data_read_dword(offs_t offset) { return (*address_space[ADDRESS_SPACE_DATA].accessors->read_dword)(offset); }
INLINE data64_t data_read_qword(offs_t offset) { return (*address_space[ADDRESS_SPACE_DATA].accessors->read_qword)(offset); }

INLINE void	data_write_byte (offs_t offset, data8_t  data) { (*address_space[ADDRESS_SPACE_DATA].accessors->write_byte)(offset, data); }
INLINE void	data_write_word (offs_t offset, data16_t data) { (*address_space[ADDRESS_SPACE_DATA].accessors->write_word)(offset, data); }
INLINE void	data_write_dword(offs_t offset, data32_t data) { (*address_space[ADDRESS_SPACE_DATA].accessors->write_dword)(offset, data); }
INLINE void	data_write_qword(offs_t offset, data64_t data) { (*address_space[ADDRESS_SPACE_DATA].accessors->write_qword)(offset, data); }

INLINE data8_t  io_read_byte (offs_t offset) { return (*address_space[ADDRESS_SPACE_IO].accessors->read_byte)(offset); }
INLINE data16_t io_read_word (offs_t offset) { return (*address_space[ADDRESS_SPACE_IO].accessors->read_word)(offset); }
INLINE data32_t io_read_dword(offs_t offset) { return (*address_space[ADDRESS_SPACE_IO].accessors->read_dword)(offset); }
INLINE data64_t io_read_qword(offs_t offset) { return (*address_space[ADDRESS_SPACE_IO].accessors->read_qword)(offset); }

INLINE void	io_write_byte (offs_t offset, data8_t  data) { (*address_space[ADDRESS_SPACE_IO].accessors->write_byte)(offset, data); }
INLINE void	io_write_word (offs_t offset, data16_t data) { (*address_space[ADDRESS_SPACE_IO].accessors->write_word)(offset, data); }
INLINE void	io_write_dword(offs_t offset, data32_t data) { (*address_space[ADDRESS_SPACE_IO].accessors->write_dword)(offset, data); }
INLINE void	io_write_qword(offs_t offset, data64_t data) { (*address_space[ADDRESS_SPACE_IO].accessors->write_qword)(offset, data); }

/* ----- safe opcode and opcode argument reading ----- */
data8_t		cpu_readop_safe(offs_t offset);
data16_t	cpu_readop16_safe(offs_t offset);
data32_t	cpu_readop32_safe(offs_t offset);
data64_t	cpu_readop64_safe(offs_t offset);
data8_t		cpu_readop_arg_safe(offs_t offset);
data16_t	cpu_readop_arg16_safe(offs_t offset);
data32_t	cpu_readop_arg32_safe(offs_t offset);
data64_t	cpu_readop_arg64_safe(offs_t offset);

/* ----- unsafe opcode and opcode argument reading ----- */
#define cpu_readop_unsafe(A)		(OP_ROM[(A) & address_space[ADDRESS_SPACE_PROGRAM].addrmask])
#define cpu_readop16_unsafe(A)		(*(data16_t *)&OP_ROM[(A) & address_space[ADDRESS_SPACE_PROGRAM].addrmask])
#define cpu_readop32_unsafe(A)		(*(data32_t *)&OP_ROM[(A) & address_space[ADDRESS_SPACE_PROGRAM].addrmask])
#define cpu_readop64_unsafe(A)		(*(data64_t *)&OP_ROM[(A) & address_space[ADDRESS_SPACE_PROGRAM].addrmask])
#define cpu_readop_arg_unsafe(A)	(OP_RAM[(A) & address_space[ADDRESS_SPACE_PROGRAM].addrmask])
#define cpu_readop_arg16_unsafe(A)	(*(data16_t *)&OP_RAM[(A) & address_space[ADDRESS_SPACE_PROGRAM].addrmask])
#define cpu_readop_arg32_unsafe(A)	(*(data32_t *)&OP_RAM[(A) & address_space[ADDRESS_SPACE_PROGRAM].addrmask])
#define cpu_readop_arg64_unsafe(A)	(*(data64_t *)&OP_RAM[(A) & address_space[ADDRESS_SPACE_PROGRAM].addrmask])

/* ----- opcode and opcode argument reading ----- */
INLINE data8_t  cpu_readop(offs_t A)		{ if (address_is_unsafe(A)) { memory_set_opbase(A); } return cpu_readop_unsafe(A); }
INLINE data16_t cpu_readop16(offs_t A)		{ if (address_is_unsafe(A)) { memory_set_opbase(A); } return cpu_readop16_unsafe(A); }
INLINE data32_t cpu_readop32(offs_t A)		{ if (address_is_unsafe(A)) { memory_set_opbase(A); } return cpu_readop32_unsafe(A); }
INLINE data64_t cpu_readop64(offs_t A)		{ if (address_is_unsafe(A)) { memory_set_opbase(A); } return cpu_readop64_unsafe(A); }
INLINE data8_t  cpu_readop_arg(offs_t A)	{ if (address_is_unsafe(A)) { memory_set_opbase(A); } return cpu_readop_arg_unsafe(A); }
INLINE data16_t cpu_readop_arg16(offs_t A)	{ if (address_is_unsafe(A)) { memory_set_opbase(A); } return cpu_readop_arg16_unsafe(A); }
INLINE data32_t cpu_readop_arg32(offs_t A)	{ if (address_is_unsafe(A)) { memory_set_opbase(A); } return cpu_readop_arg32_unsafe(A); }
INLINE data64_t cpu_readop_arg64(offs_t A)	{ if (address_is_unsafe(A)) { memory_set_opbase(A); } return cpu_readop_arg64_unsafe(A); }

/* ----- bank switching for CPU cores ----- */
#define change_pc(pc)																	\
do {																					\
	if (address_space[ADDRESS_SPACE_PROGRAM].readlookup[LEVEL1_INDEX((pc) & address_space[ADDRESS_SPACE_PROGRAM].addrmask)] != opcode_entry)	\
		memory_set_opbase(pc);															\
} while (0)																				\

/* ----- forces the next branch to generate a call to the opbase handler ----- */
#define catch_nextBranch()			(opcode_entry = 0xff)

/* ----- bank switching macro ----- */
#define cpu_setbank(bank, base) 														\
do {																					\
	if (bank >= STATIC_BANK1 && bank <= STATIC_BANKMAX)									\
	{																					\
		cpu_bankbase[bank] = (UINT8 *)(base);											\
		if (opcode_entry == bank && cpu_getactivecpu() >= 0)							\
		{																				\
			opcode_entry = 0xff;														\
			memory_set_opbase(activecpu_get_pc_byte());									\
		}																				\
	}																					\
} while (0)



/***************************************************************************

	Backwards compatibility

***************************************************************************/

#define install_mem_read_handler(c,s,e,h)		memory_install_read8_handler(c,ADDRESS_SPACE_PROGRAM,s,e,0,h)
#define install_mem_read16_handler(c,s,e,h)		memory_install_read16_handler(c,ADDRESS_SPACE_PROGRAM,s,e,0,h)
#define install_mem_read32_handler(c,s,e,h)		memory_install_read32_handler(c,ADDRESS_SPACE_PROGRAM,s,e,0,h)
#define install_mem_read64_handler(c,s,e,h)		memory_install_read64_handler(c,ADDRESS_SPACE_PROGRAM,s,e,0,h)
#define install_mem_write_handler(c,s,e,h)		memory_install_write8_handler(c,ADDRESS_SPACE_PROGRAM,s,e,0,h)
#define install_mem_write16_handler(c,s,e,h)	memory_install_write16_handler(c,ADDRESS_SPACE_PROGRAM,s,e,0,h)
#define install_mem_write32_handler(c,s,e,h)	memory_install_write32_handler(c,ADDRESS_SPACE_PROGRAM,s,e,0,h)
#define install_mem_write64_handler(c,s,e,h)	memory_install_write64_handler(c,ADDRESS_SPACE_PROGRAM,s,e,0,h)

#define install_port_read_handler(c,s,e,h)		memory_install_read8_handler(c,ADDRESS_SPACE_IO,s,e,0,h)
#define install_port_read16_handler(c,s,e,h)	memory_install_read16_handler(c,ADDRESS_SPACE_IO,s,e,0,h)
#define install_port_read32_handler(c,s,e,h)	memory_install_read32_handler(c,ADDRESS_SPACE_IO,s,e,0,h)
#define install_port_read64_handler(c,s,e,h)	memory_install_read64_handler(c,ADDRESS_SPACE_IO,s,e,0,h)
#define install_port_write_handler(c,s,e,h)		memory_install_write8_handler(c,ADDRESS_SPACE_IO,s,e,0,h)
#define install_port_write16_handler(c,s,e,h)	memory_install_write16_handler(c,ADDRESS_SPACE_IO,s,e,0,h)
#define install_port_write32_handler(c,s,e,h)	memory_install_write32_handler(c,ADDRESS_SPACE_IO,s,e,0,h)
#define install_port_write64_handler(c,s,e,h)	memory_install_write64_handler(c,ADDRESS_SPACE_IO,s,e,0,h)

#define READ_HANDLER READ8_HANDLER
#define WRITE_HANDLER WRITE8_HANDLER

#ifdef __cplusplus
}
#endif

#endif	/* !_MEMORY_H */

